"""Tests publish_dataset.py."""

import os
import shutil
from mlops.examples.image.classification.publish_dataset import publish_dataset

TEST_PUBLICATION_PATH_LOCAL = '/tmp/test_publish_dataset/datasets'


def test_publish_dataset_creates_files() -> None:
    """Tests that publish_dataset creates the dataset files."""
    try:
        shutil.rmtree(TEST_PUBLICATION_PATH_LOCAL)
    except FileNotFoundError:
        pass
    publish_dataset(TEST_PUBLICATION_PATH_LOCAL)
    assert os.path.exists(TEST_PUBLICATION_PATH_LOCAL)



"""Contains the VersionedDataset class."""

import os
import json
import dill as pickle
import numpy as np
from s3fs import S3FileSystem


class VersionedDataset:
    """Represents a versioned dataset."""

    def __init__(self, path: str) -> None:
        """Instantiates the object.

        :param path: The path, either on the local filesystem or in a cloud
            store such as S3, from which the dataset should be loaded. An S3
            path should be a URL of the form "s3://bucket-name/path/to/dir".
        """
        self.path = path
        if path.startswith('s3://'):
            fs = S3FileSystem()
            # Get tensors.
            tensor_paths = {tensor_path
                            for tensor_path in fs.ls(path)
                            if tensor_path.endswith('.npy')}
            for tensor_path in tensor_paths:
                attr_name = tensor_path.split('.npy')[0].split('/')[-1]
                with fs.open(tensor_path, 'rb') as infile:
                    tensor = np.load(infile)
                setattr(self, attr_name, tensor)
            # Get hash.
            with fs.open(os.path.join(path, 'meta.json'),
                         'r',
                         encoding='utf-8') as infile:
                metadata = json.loads(infile.read())
            self.md5 = metadata['hash']
            # Get data processor.
            with fs.open(os.path.join(path, 'data_processor.pkl'),
                         'rb') as infile:
                processor = pickle.loads(infile.read(), ignore=True)
            self.data_processor = processor
        else:
            # Get tensors.
            tensor_filenames = {tensor_filename
                                for tensor_filename in os.listdir(path)
                                if tensor_filename.endswith('.npy')}
            for tensor_filename in tensor_filenames:
                tensor_path = os.path.join(path, tensor_filename)
                attr_name = tensor_filename.split('.npy')[0]
                tensor = np.load(tensor_path)
                setattr(self, attr_name, tensor)
            # Get hash.
            with open(os.path.join(path, 'meta.json'),
                      'r',
                      encoding='utf-8') as infile:
                metadata = json.loads(infile.read())
            self.md5 = metadata['hash']
            # Get data processor.
            with open(os.path.join(path, 'data_processor.pkl'), 'rb') as infile:
                processor = pickle.loads(infile.read(), ignore=True)
            self.data_processor = processor

    def __eq__(self, other: 'VersionedDataset') -> bool:
        """Returns True if the two objects have the same loaded MD5 hash code,
        False otherwise.

        :param other: The dataset with which to compare this object.
        :return: True if the object MD5 hashes match.
        """
        return self.md5 == other.md5

    def __hash__(self) -> int:
        """Returns this object's hashcode based on the loaded MD5 hashcode.

        :return: The object's hashcode based on the loaded MD5 hashcode.
        """
        return hash(self.md5)



"""Tests pokemon_classification_data_processor.py."""

import pytest
import numpy as np
from mlops.examples.image.classification.pokemon_classification_data_processor \
    import PokemonClassificationDataProcessor, \
    DEFAULT_DATASET_TRAINVALTEST_PATH, DEFAULT_DATASET_PRED_PATH, \
    HEIGHT, WIDTH, CHANNELS, CLASSES
from mlops.examples.image.classification.errors import LabelsNotFoundError

EXPECTED_NUM_TRAINVALTEST = 10
EXPECTED_NUM_TRAIN = 7
EXPECTED_NUM_VAL = 2
EXPECTED_NUM_PRED = 3
PIXEL_MIN = 0
PIXEL_MAX = 255
BULBASAUR_IMG_MEAN = 0.06437409
BULBASAUR_LABEL = {'Grass', 'Poison'}
CHARIZARD_IMG_MEAN = 0.17114125
CHARIZARD_LABEL = {'Fire', 'Flying'}


def test_get_raw_features_and_labels_returns_expected_keys() -> None:
    """Tests that get_raw_features_and_labels returns the expected keys for the
    train/val/test dataset."""
    processor = PokemonClassificationDataProcessor()
    features, labels = processor.get_raw_features_and_labels(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    assert set(features.keys()) == {'X_train', 'X_val', 'X_test'}
    assert set(labels.keys()) == {'y_train', 'y_val', 'y_test'}


def test_get_raw_features_and_labels_pred_raises_error() -> None:
    """Tests that get_raw_features_and_labels raises LabelsNotFoundError when
    called on the prediction directory."""
    processor = PokemonClassificationDataProcessor()
    with pytest.raises(LabelsNotFoundError):
        _ = processor.get_raw_features_and_labels(DEFAULT_DATASET_PRED_PATH)


def test_get_raw_features_and_labels_trainvaltest_correct_split() -> None:
    """Tests that the train/val/test datasets are split into the expected
    sizes."""
    processor = PokemonClassificationDataProcessor()
    features, labels = processor.get_raw_features_and_labels(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    num_examples = sum(map(len, features.values()))
    assert num_examples == EXPECTED_NUM_TRAINVALTEST
    assert len(features['X_train']) == EXPECTED_NUM_TRAIN
    assert len(features['X_val']) == EXPECTED_NUM_VAL
    assert len(features['X_test']) == EXPECTED_NUM_TRAINVALTEST - \
           EXPECTED_NUM_TRAIN - EXPECTED_NUM_VAL
    assert len(features['X_train']) == len(labels['y_train'])
    assert len(features['X_val']) == len(labels['y_val'])
    assert len(features['X_test']) == len(labels['y_test'])


def test_get_raw_features_trainvaltest_returns_expected_keys() -> None:
    """Tests that get_raw_features returns the expected keys {'X_train',
    'X_val', 'X_test} when called on the train/val/test directory.
    """
    processor = PokemonClassificationDataProcessor()
    raw = processor.get_raw_features(DEFAULT_DATASET_TRAINVALTEST_PATH)
    assert set(raw.keys()) == {'X_train', 'X_val', 'X_test'}


def test_get_raw_features_match() -> None:
    """Tests that the features produced by get_raw_features_and_labels and
    get_raw_features are the same features."""
    # pylint: disable=invalid-name
    processor = PokemonClassificationDataProcessor()
    features, _ = processor.get_raw_features_and_labels(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    X_all = np.concatenate((features['X_train'],
                            features['X_val'],
                            features['X_test']))
    features_only = processor.get_raw_features(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    X_all_only = np.concatenate((features_only['X_train'],
                                 features_only['X_val'],
                                 features_only['X_test']))
    X_all.sort(axis=0)
    X_all_only.sort(axis=0)
    assert np.array_equal(X_all, X_all_only)


def test_get_raw_features_pred_returns_expected_keys() -> None:
    """Tests that get_raw_features returns the expected keys {'X_pred'} when
    called on the prediction directory.
    """
    processor = PokemonClassificationDataProcessor()
    raw = processor.get_raw_features(DEFAULT_DATASET_PRED_PATH)
    assert set(raw.keys()) == {'X_pred'}


def test_get_raw_features_correct_shape() -> None:
    """Tests that get_raw_features returns tensors with the expected shapes."""
    processor = PokemonClassificationDataProcessor()
    raw = processor.get_raw_features(DEFAULT_DATASET_PRED_PATH)
    for tensor in raw.values():
        assert tensor.shape[1:] == (HEIGHT, WIDTH, CHANNELS)


def test_get_raw_features_correct_dtype() -> None:
    """Tests that get_raw_features returns tensors with dtype float32."""
    processor = PokemonClassificationDataProcessor()
    raw = processor.get_raw_features(DEFAULT_DATASET_PRED_PATH)
    for tensor in raw.values():
        assert tensor.dtype == np.float32


def test_get_raw_features_correct_value_range() -> None:
    """Tests that get_raw_features returns tensors in the range [0, 255]."""
    processor = PokemonClassificationDataProcessor()
    raw = processor.get_raw_features(DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        assert tensor.min() >= 0
        assert tensor.max() <= 1


def test_get_raw_features_no_na() -> None:
    """Tests that get_raw_features returns tensors with no missing values."""
    processor = PokemonClassificationDataProcessor()
    raw = processor.get_raw_features(DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        assert not np.isnan(tensor).any()


def test_get_raw_features_have_multiple_pixel_values() -> None:
    """Tests that the images were loaded correctly by ensuring that more than
    one pixel value exists in the tensors."""
    processor = PokemonClassificationDataProcessor()
    raw = processor.get_raw_features(DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        assert len(np.unique(tensor)) > 1


def test_get_raw_labels_trainvaltest_lengths_match_features() -> None:
    """Tests that all entries in the raw label dictionary have the same number
    of examples as their counterpart features."""
    processor = PokemonClassificationDataProcessor()
    raw_features, raw_labels = processor.get_raw_features_and_labels(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    assert len(raw_features['X_train']) == len(raw_labels['y_train'])
    assert len(raw_features['X_val']) == len(raw_labels['y_val'])
    assert len(raw_features['X_test']) == len(raw_labels['y_test'])


def test_get_raw_labels_correct_tensor_shapes() -> None:
    """Tests that labels are of the correct shape."""
    processor = PokemonClassificationDataProcessor()
    _, raw = processor.get_raw_features_and_labels(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        assert tensor.shape[1:] == (2,)


def test_get_raw_labels_correct_dtype() -> None:
    """Tests that labels are of type object (string)."""
    processor = PokemonClassificationDataProcessor()
    _, raw = processor.get_raw_features_and_labels(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        assert tensor.dtype == object


def test_get_raw_labels_valid_classes() -> None:
    """Tests that all raw label classes are valid Pokemon types."""
    processor = PokemonClassificationDataProcessor()
    _, raw = processor.get_raw_features_and_labels(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        for row in tensor:
            assert row[0] in CLASSES
            assert row[1] is None or row[1] in CLASSES


def test_preprocessed_features_same_shape_as_raw() -> None:
    """Tests that the preprocessed features have the same shape as the raw
    features."""
    processor = PokemonClassificationDataProcessor()
    raw = processor.get_raw_features(DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        preprocessed = processor.preprocess_features(tensor)
        assert tensor.shape == preprocessed.shape


def test_preprocess_features_correct_dtype() -> None:
    """Tests that preprocessed features are of dtype float32."""
    processor = PokemonClassificationDataProcessor()
    raw = processor.get_raw_features(DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        preprocessed = processor.preprocess_features(tensor)
        assert preprocessed.dtype == np.float32


def test_preprocess_features_no_na() -> None:
    """Tests that preprocessed features have no missing values."""
    processor = PokemonClassificationDataProcessor()
    raw = processor.get_raw_features(DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        preprocessed = processor.preprocess_features(tensor)
        assert not np.isnan(preprocessed).any()


def test_preprocessed_features_scaled() -> None:
    """Tests that preprocessing scales the features to the range [0, 1]."""
    processor = PokemonClassificationDataProcessor()
    raw = processor.get_raw_features(DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        preprocessed = processor.preprocess_features(tensor)
        assert preprocessed.min() >= 0
        assert preprocessed.max() <= 1


def test_preprocess_labels_correct_shape() -> None:
    """Tests that the preprocessed labels have the correct shape."""
    processor = PokemonClassificationDataProcessor()
    _, raw = processor.get_raw_features_and_labels(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        preprocessed = processor.preprocess_labels(tensor)
        assert preprocessed.shape == (len(tensor), len(CLASSES))


def test_preprocess_labels_correct_dtype() -> None:
    """Tests that the preprocessed labels are of dtype float32."""
    processor = PokemonClassificationDataProcessor()
    _, raw = processor.get_raw_features_and_labels(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        preprocessed = processor.preprocess_labels(tensor)
        assert preprocessed.dtype == np.float32


def test_preprocess_labels_no_na() -> None:
    """Tests that the preprocessed labels have no missing values."""
    processor = PokemonClassificationDataProcessor()
    _, raw = processor.get_raw_features_and_labels(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        preprocessed = processor.preprocess_labels(tensor)
        assert not np.isnan(preprocessed).any()


def test_preprocess_labels_binary() -> None:
    """Tests that the preprocessed labels have values in the set {0, 1}."""
    processor = PokemonClassificationDataProcessor()
    _, raw = processor.get_raw_features_and_labels(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        preprocessed = processor.preprocess_labels(tensor)
        assert set(np.unique(preprocessed)) == {0, 1}


def test_preprocess_labels_min_one_max_two_classes() -> None:
    """Tests that each preprocessed label has at least one and at most two
    ones indicating the class(es)."""
    processor = PokemonClassificationDataProcessor()
    _, raw = processor.get_raw_features_and_labels(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        preprocessed = processor.preprocess_labels(tensor)
        row_sums = preprocessed.sum(axis=1)
        assert set(np.unique(row_sums)).union({1, 2}) == {1, 2}


def test_unpreprocess_features_inverts_transformation() -> None:
    """Tests that unpreprocessing the preprocessed features results in the raw
    features."""
    processor = PokemonClassificationDataProcessor()
    raw = processor.get_raw_features(DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        preprocessed = processor.preprocess_features(tensor)
        unpreprocessed = processor.unpreprocess_features(preprocessed)
        assert (unpreprocessed == tensor).all()


def test_unpreprocess_labels_inverts_transformation() -> None:
    """Tests that unpreprocessing the preprocessed labels results in the raw
    labels."""
    processor = PokemonClassificationDataProcessor()
    _, raw = processor.get_raw_features_and_labels(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    for tensor in raw.values():
        preprocessed = processor.preprocess_labels(tensor)
        unpreprocessed = processor.unpreprocess_labels(preprocessed)
        assert (unpreprocessed == tensor).all()


def test_get_raw_features_and_labels_examples_in_same_order() -> None:
    """Tests that the raw features and raw labels have examples in the same
    order. For example, say X_train[0] is the raw Bulbasaur image; then
    y_train[0] must be the labels for Bulbasaur."""
    # pylint: disable=invalid-name
    processor = PokemonClassificationDataProcessor()
    features, labels = processor.get_raw_features_and_labels(
        DEFAULT_DATASET_TRAINVALTEST_PATH)
    X_all = np.concatenate((features['X_train'],
                            features['X_val'],
                            features['X_test']))
    y_all = np.concatenate((labels['y_train'],
                            labels['y_val'],
                            labels['y_test']))
    bulbasaur_idx = None
    for idx, arr in enumerate(X_all):
        if np.isclose(arr.mean(), BULBASAUR_IMG_MEAN):
            bulbasaur_idx = idx
    assert bulbasaur_idx is not None
    assert set(y_all[bulbasaur_idx]) == BULBASAUR_LABEL
    charizard_idx = None
    for idx, arr in enumerate(X_all):
        if np.isclose(arr.mean(), CHARIZARD_IMG_MEAN):
            charizard_idx = idx
    assert charizard_idx is not None
    assert set(y_all[charizard_idx]) == CHARIZARD_LABEL


def test_get_valid_prediction_correct_shape() -> None:
    """Tests that the output of get_valid_prediction is of the same shape as
    the input."""
    pred_arr = np.array([[0.8, 0.4, 0.2, 0.6],
                         [0.3, 0.4, 0.1, 0.1]])
    valid = PokemonClassificationDataProcessor.get_valid_prediction(pred_arr)
    assert pred_arr.shape == valid.shape


def test_get_valid_prediction_output_is_binary() -> None:
    """Tests that the output of get_valid_prediction on arbitrary input is
    binary."""
    pred_arr = np.array([[0.8, 0.4, 0.2, 0.6],
                         [0.3, 0.4, 0.1, 0.1],
                         [-1, 5, 2, 0.5]])
    valid = PokemonClassificationDataProcessor.get_valid_prediction(pred_arr)
    assert set(np.unique(valid)) == {0, 1}


def test_get_valid_prediction_chooses_highest() -> None:
    """Tests that get_valid_prediction chooses the highest scores as output."""
    pred_arr = np.array([[0.8, 0.4, 0.2, 0.6],
                         [0.3, 0.4, 0.1, 0.1],
                         [0.9, 0.9, 0.8, 0.8]])
    valid = PokemonClassificationDataProcessor.get_valid_prediction(pred_arr)
    assert valid.tolist() == [[1, 0, 0, 1],
                              [0, 1, 0, 0],
                              [1, 1, 0, 0]]


def test_get_valid_prediction_one_or_two_classes() -> None:
    """Tests that get_valid_prediction returns predictions with one or two
    classes."""
    pred_arr = np.array([[0.8, 0.4, 0.2, 0.6],
                         [0.3, 0.4, 0.1, 0.1],
                         [0.9, 0.9, 0.9, 0.9],
                         [0.1, 0.1, 0.1, 0.1],
                         [2.0, 2.0, 2.0, 2.0]])
    valid = PokemonClassificationDataProcessor.get_valid_prediction(pred_arr)
    row_sums = valid.sum(axis=1)
    assert set(row_sums) == {1, 2}


def test_get_valid_prediction_threshold_only_affects_second_highest() -> None:
    """Tests that the decision threshold only affects the second highest
    prediction value."""
    pred_arr = np.array([[0.8, 0.4, 0.2, 0.6],
                         [0.3, 0.4, 0.1, 0.1],
                         [0.9, 0.8, 0.7, 0.7]])
    valid = PokemonClassificationDataProcessor.get_valid_prediction(
        pred_arr, threshold=0.6)
    assert valid.tolist() == [[1, 0, 0, 1],
                              [0, 1, 0, 0],
                              [1, 1, 0, 0]]
    valid = PokemonClassificationDataProcessor.get_valid_prediction(
        pred_arr, threshold=0.99)
    assert valid.tolist() == [[1, 0, 0, 0],
                              [0, 1, 0, 0],
                              [1, 0, 0, 0]]



"""Tests model_prediction.py."""
# pylint: disable=no-name-in-module,no-member

import os
import json
import pytest
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Flatten
from mlops.dataset.versioned_dataset import VersionedDataset
from mlops.model.versioned_model import VersionedModel
from mlops.examples.image.classification.publish_dataset import DATASET_VERSION
from mlops.examples.image.classification.train_model import train_model, \
    publish_model
from mlops.examples.image.classification import model_prediction
from mlops.examples.image.classification.pokemon_classification_data_processor \
    import DEFAULT_DATASET_PRED_PATH
from mlops.examples.image.classification.errors import NoModelPathsSuppliedError
from tests.examples.image.classification.test_train_model import \
    _create_dataset, TEST_DATASET_PUBLICATION_PATH_LOCAL

TEST_MODEL_PUBLICATION_PATH_LOCAL = ('/tmp/test_model_prediction/models/'
                                     'versioned')
EXPECTED_PRED_SHAPE = (3, 2)
FAKE_TRAIN_HISTORIES = [{'loss': [1, 2, 3], 'val_loss': [4, 5, 6]},
                        {'loss': [9, 5, 1], 'val_loss': [7, 7, 7]},
                        {'loss': [2, 2, 2], 'val_loss': [4, 5, 1]}]
BEST_VAL_IDX = 2


def _create_model() -> str:
    """Creates a test model.

    :return: The model's publication path.
    """
    dataset = VersionedDataset(os.path.join(TEST_DATASET_PUBLICATION_PATH_LOCAL,
                                            DATASET_VERSION))
    model = Sequential([
        Flatten(input_shape=dataset.X_train.shape[1:]),
        Dense(dataset.y_train.shape[1])])
    model.compile('adam', loss='mse')
    train_kwargs = {'epochs': 3, 'batch_size': 8}
    training_config = train_model(
        model,
        dataset,
        **train_kwargs)
    return publish_model(model,
                         dataset,
                         training_config,
                         TEST_MODEL_PUBLICATION_PATH_LOCAL)


def test_model_evaluate_returns_valid_loss() -> None:
    """Tests that model evaluation returns a loss value that is valid."""
    _create_dataset()
    publication_path = _create_model()
    dataset = VersionedDataset(os.path.join(TEST_DATASET_PUBLICATION_PATH_LOCAL,
                                            DATASET_VERSION))
    model = VersionedModel(publication_path)
    assert model_prediction.model_evaluate(dataset, model) >= 0


def test_model_predict_correct_shape() -> None:
    """Tests that model predictions are of the correct shape."""
    _create_dataset()
    publication_path = _create_model()
    dataset = VersionedDataset(os.path.join(TEST_DATASET_PUBLICATION_PATH_LOCAL,
                                            DATASET_VERSION))
    model = VersionedModel(publication_path)
    features = dataset.data_processor.get_preprocessed_features(
        DEFAULT_DATASET_PRED_PATH)['X_pred']
    predictions = model_prediction.model_predict(features, dataset, model)
    assert predictions.shape == EXPECTED_PRED_SHAPE


def test_get_best_model_with_no_models_raises_error() -> None:
    """Tests that get_best_model raises an error when an empty path collection
    is supplied."""
    with pytest.raises(NoModelPathsSuppliedError):
        _ = model_prediction.get_best_model([])


def test_get_best_model_gets_lowest_val_error() -> None:
    """Tests that get_best_model gets the model with the lowest validation
    error."""
    _create_dataset()
    publication_paths = []
    for fake_train_history in FAKE_TRAIN_HISTORIES:
        publication_path = _create_model()
        publication_paths.append(publication_path)
        with open(os.path.join(publication_path, 'meta.json'), 'r',
                  encoding='utf-8') as infile:
            contents = json.loads(infile.read())
        contents['history'] = fake_train_history
        with open(os.path.join(publication_path, 'meta.json'), 'w',
                  encoding='utf-8') as outfile:
            outfile.write(json.dumps(contents))
    best_model = model_prediction.get_best_model(publication_paths)
    assert best_model.path == publication_paths[BEST_VAL_IDX]



"""Tests invertible_data_processor.py."""

import pytest
import numpy as np
from mlops.dataset import invertible_data_processor as idp
from tests.dataset.preset_data_processor import PresetDataProcessor


@pytest.mark.xfail
def test_invertible_data_processor_is_abstract() -> None:
    """Tests that the InvertibleDataProcessor object is abstract."""
    # This test currently fails because we needed to make
    # InvertibleDataProcessor concrete as a serialization workaround. See:
    # https://github.com/kostaleonard/mlops/issues/12
    with pytest.raises(TypeError):
        _ = idp.InvertibleDataProcessor()


def test_unpreprocess_features_returns_raw_features() -> None:
    """Tests that unpreprocess_features inverts feature preprocessing."""
    processor = PresetDataProcessor()
    features = processor.get_preprocessed_features('path/to/my/features')
    raw_features = processor.get_raw_features('path/to/my/features')
    for name in features:
        inverted_features = processor.unpreprocess_features(features[name])
        assert np.array_equal(inverted_features, raw_features[name])


def test_unpreprocess_labels_returns_raw_labels() -> None:
    """Tests that unpreprocess_labels inverts label preprocessing."""
    processor = PresetDataProcessor()
    _, labels = processor.get_preprocessed_features_and_labels(
        'path/to/my/dataset')
    _, raw_labels = processor.get_raw_features_and_labels(
        'path/to/my/dataset')
    for name in labels:
        inverted_labels = processor.unpreprocess_labels(labels[name])
        assert np.array_equal(inverted_labels, raw_labels[name])



"""Trains a new model on the Pokemon classification task."""
# pylint: disable=no-name-in-module

import os
from typing import Optional, Any
from tensorflow.keras.models import Model, Sequential
from tensorflow.keras.layers import Conv2D, MaxPooling2D, Dense, Flatten, \
    Dropout
from tensorflow.keras.callbacks import ModelCheckpoint
from mlops.errors import PublicationPathAlreadyExistsError
from mlops.dataset.versioned_dataset import VersionedDataset
from mlops.model.versioned_model_builder import VersionedModelBuilder
from mlops.model.training_config import TrainingConfig
from mlops.examples.image.classification.publish_dataset import \
    publish_dataset, DATASET_PUBLICATION_PATH_LOCAL, DATASET_VERSION

MODEL_PUBLICATION_PATH_LOCAL = os.path.join('models', 'pokemon', 'versioned')
MODEL_PUBLICATION_PATH_S3 = 's3://kosta-mlops/models/pokemon'
MODEL_CHECKPOINT_FILENAME = os.path.join('models', 'pokemon', 'checkpoints',
                                         'model_best.h5')
TAGS = ['baseline']


def get_baseline_model(dataset: VersionedDataset) -> Model:
    """Returns a new Keras Model for use on the dataset. This model is only a
    baseline; developers should also experiment with custom models in notebook
    environments.

    :param dataset: The input dataset. Used to determine model input and output
        shapes.
    :return: A new Keras Model for use on the dataset.
    """
    model = Sequential()
    # Shape: (None, 120, 120, 3).
    model.add(Conv2D(16, (3, 3), activation='relu',
                     input_shape=dataset.X_train.shape[1:]))
    # Shape: (None, 118, 118, 16).
    model.add(MaxPooling2D((2, 2)))
    # Shape: (None, 59, 59, 16).
    model.add(Conv2D(32, (3, 3), activation='relu'))
    # Shape: (None, 57, 57, 32).
    model.add(MaxPooling2D((2, 2)))
    # Shape: (None, 28, 28, 32).
    model.add(Conv2D(64, (3, 3), activation='relu'))
    # Shape: (None, 26, 26, 64).
    model.add(MaxPooling2D((2, 2)))
    # Shape: (None, 13, 13, 64).
    model.add(Conv2D(128, (3, 3), activation='relu'))
    # Shape: (None, 11, 11, 128).
    model.add(MaxPooling2D((2, 2)))
    # Shape: (None, 5, 5, 128).
    model.add(Conv2D(256, (3, 3), activation='relu'))
    # Shape: (None, 3, 3, 256).
    model.add(Flatten())
    # Shape: (None, 2304).
    model.add(Dense(128, activation='relu'))
    # Shape: (None, 128).
    model.add(Dropout(0.4))
    # Shape: (None, 128).
    model.add(Dense(dataset.y_train.shape[1], activation='sigmoid'))
    # Shape: (None, 18).
    model.compile(optimizer='adam', loss='binary_crossentropy')
    return model


def train_model(model: Model,
                dataset: VersionedDataset,
                model_checkpoint_filename: Optional[str] = None,
                **fit_kwargs: Any) -> TrainingConfig:
    """Trains the model on the dataset and returns the training configuration
    object.

    :param model: The Keras Model to be trained.
    :param dataset: The input dataset.
    :param model_checkpoint_filename: If supplied, saves model checkpoints to
        the specified path.
    :param fit_kwargs: Keyword arguments to be passed to model.fit().
    :return: The training configuration.
    """
    callbacks = []
    if model_checkpoint_filename:
        checkpoint_callback = ModelCheckpoint(
            model_checkpoint_filename,
            save_best_only=True)
        callbacks.append(checkpoint_callback)
    history = model.fit(x=dataset.X_train,
                        y=dataset.y_train,
                        validation_data=(dataset.X_val, dataset.y_val),
                        callbacks=callbacks,
                        **fit_kwargs)
    return TrainingConfig(history, fit_kwargs)


def publish_model(model: Model,
                  dataset: VersionedDataset,
                  training_config: TrainingConfig,
                  publication_path: str,
                  tags: Optional[list[str]] = None) -> str:
    """Publishes the model to the path on the local or remote filesystem.

    :param model: The model to be published, with the exact weights desired for
        publication (the user needs to set the weights to the best found during
        training if that is what they desire).
    :param dataset: The input dataset.
    :param training_config: The training configuration.
    :param publication_path: The path to which the model will be published.
    :param tags: Optional tags for the published model.
    :return: The versioned model's publication path.
    """
    builder = VersionedModelBuilder(dataset, model, training_config)
    return builder.publish(publication_path, tags=tags)


def main() -> None:
    """Runs the program."""
    try:
        dataset_path = publish_dataset(DATASET_PUBLICATION_PATH_LOCAL)
    except PublicationPathAlreadyExistsError:
        dataset_path = os.path.join(DATASET_PUBLICATION_PATH_LOCAL,
                                    DATASET_VERSION)
    dataset = VersionedDataset(dataset_path)
    model = get_baseline_model(dataset)
    training_config = train_model(
        model,
        dataset,
        model_checkpoint_filename=MODEL_CHECKPOINT_FILENAME,
        epochs=5,
        batch_size=4)
    publish_model(
        model,
        dataset,
        training_config,
        MODEL_PUBLICATION_PATH_LOCAL,
        tags=TAGS)


if __name__ == '__main__':
    main()



"""mlops is a package for conducting MLOps, including versioning of datasets and
models."""

__version__ = '0.0.11'

import mlops.dataset
import mlops.errors
import mlops.model

__all__ = [
    '__version__',
    'dataset',
    'errors',
    'model'
]



import dill as pickle
class A:
    def foo(self):
        return 'original'

a = A()
obj_type = type(a)
obj_module = obj_type.__module__
try:
    obj_type.__module__ = '__main__'
    serialized = pickle.dumps(a)
finally:
    obj_type.__module__ = obj_module

class A:
    def foo(self):
        return 'new'

print(pickle.loads(serialized, ignore=True).foo())


if __name__ == '__main__':
    pass



"""Contains the VersionedDatasetBuilder class."""

import os
from pathlib import Path
import shutil
from tempfile import TemporaryFile
from typing import Optional
from datetime import datetime
import json
import dill as pickle
import numpy as np
from s3fs import S3FileSystem
from mlops.dataset.data_processor import DataProcessor
from mlops.hashing.hashing import get_hash_local, get_hash_s3
from mlops.errors import PublicationPathAlreadyExistsError, \
    InvalidDatasetCopyStrategyError

STRATEGY_COPY = 'copy'
STRATEGY_LINK = 'link'


class VersionedDatasetBuilder:
    """An object containing all of the components that form a versioned dataset.
    This object is only used to ensure a standard format for datasets stored in
    a dataset archive (such as the local filesystem or S3), and is not meant for
    consumption by downstream models."""
    # pylint: disable=too-few-public-methods

    def __init__(self,
                 dataset_path: str,
                 data_processor: DataProcessor) -> None:
        """Instantiates the object. Features and labels will be extracted from
        the dataset path using the DataProcessor object. When this object is
        published, the feature and label tensors will be saved with '.npy' as
        the suffix. For example, if a tensor is named 'X_train', it will be
        saved as 'X_train.npy'.

        :param dataset_path: The path to the file or directory on the local or
            remote filesystem containing the dataset.
        :param data_processor: The DataProcessor object with which the features
            and labels were generated. This object is saved so that properly
            formatted features can be generated at prediction time, and so that
            features and labels can be "unpreprocessed" to match their original
            representations.
        """
        self.dataset_path = dataset_path.rstrip('/')
        self.data_processor = data_processor
        self.features, self.labels = \
            data_processor.get_preprocessed_features_and_labels(dataset_path)

    def publish(self,
                path: str,
                version: Optional[str] = None,
                dataset_copy_strategy: str = STRATEGY_COPY,
                tags: Optional[list[str]] = None) -> str:
        """Saves the versioned dataset files to the given path. If the path and
        appended version already exists, this operation will raise a
        PublicationPathAlreadyExistsError.

        The following files will be created:
            path/version/ (the publication path and version)
                X_train.npy (and other feature tensors by their given names)
                y_train.npy (and other label tensors by their given names)
                data_processor.pkl (DataProcessor object)
                meta.json (metadata)
                raw/ (non-empty directory with the raw dataset files)
                    ...

        The contents of meta.json will be:
            {
                version: (dataset version)
                hash: (MD5 hash of all objects apart from data_processor.pkl and
                    meta.json)
                created_at: (timestamp)
                tags: (optional list of tags)
            }

        :param path: The path, either on the local filesystem or in a cloud
            store such as S3, to which the dataset should be saved. The version
            will be appended to this path as a subdirectory. An S3 path
            should be a URL of the form "s3://bucket-name/path/to/dir". It is
            recommended to use this same path to publish all datasets, since it
            will prevent the user from creating two different datasets with the
            same version.
        :param version: A string indicating the dataset version. The version
            should be unique to this dataset. If None, the publication timestamp
            will be used as the version.
        :param dataset_copy_strategy: The strategy by which to copy the
            original, raw dataset to the published path. The default is
            STRATEGY_COPY, which recursively copies all files and directories
            from the dataset path supplied at instantiation to the published
            path so that the dataset can be properly versioned. STRATEGY_LINK
            will instead create a file 'link.txt' containing the supplied
            dataset path; this is desirable if the raw dataset is already stored
            in a versioned repository, and copying would create an unnecessary
            duplicate.
        :param tags: An optional list of string tags to add to the dataset
            metadata.
        :return: The versioned dataset's publication path.
        """
        timestamp = datetime.now().isoformat()
        if not version:
            version = timestamp
        if not tags:
            tags = []
        if dataset_copy_strategy not in {STRATEGY_COPY, STRATEGY_LINK}:
            raise InvalidDatasetCopyStrategyError
        publication_path = os.path.join(path.rstrip('/'), version)
        copy_path = os.path.join(publication_path, 'raw')
        link_path = os.path.join(copy_path, 'link.txt')
        metadata_path = os.path.join(publication_path, 'meta.json')
        processor_path = os.path.join(publication_path, 'data_processor.pkl')
        metadata = {
            'version': version,
            'hash': 'TDB',
            'created_at': timestamp,
            'tags': tags}
        if path.startswith('s3://'):
            return self._publish_s3(publication_path,
                                    copy_path,
                                    link_path,
                                    metadata_path,
                                    processor_path,
                                    dataset_copy_strategy,
                                    metadata)
        return self._publish_local(publication_path,
                                   copy_path,
                                   link_path,
                                   metadata_path,
                                   processor_path,
                                   dataset_copy_strategy,
                                   metadata)

    def _publish_local(self,
                       publication_path: str,
                       copy_path: str,
                       link_path: str,
                       metadata_path: str,
                       processor_path: str,
                       dataset_copy_strategy: str,
                       metadata: dict) -> str:
        """Saves the versioned dataset files to the given local path. See
        publish() for more detailed information.

        :param publication_path: The local path to which to publish the dataset.
        :param copy_path: The path to which the raw dataset is copied.
        :param link_path: The path to the file containing the raw dataset link.
        :param metadata_path: The path to which the metadata should be saved.
        :param processor_path: The path to which to write the data processor.
        :param dataset_copy_strategy: The strategy by which to copy the
            original, raw dataset to the published path.
        :param metadata: Dataset metadata.
        :return: The versioned dataset's publication path.
        """
        # pylint: disable=too-many-arguments
        files_to_hash = set()
        # Create publication path.
        VersionedDatasetBuilder._make_publication_path_local(publication_path)
        # Save tensors.
        self._write_tensors_local(publication_path)
        # Save the raw dataset.
        if dataset_copy_strategy == STRATEGY_LINK:
            self._make_raw_dataset_link_local(link_path)
            files_to_hash.add(link_path)
        else:
            file_paths = self._copy_raw_dataset_local(copy_path)
            files_to_hash = files_to_hash.union(file_paths)
        # Save metadata.
        hash_digest = get_hash_local(files_to_hash)
        metadata['hash'] = hash_digest
        VersionedDatasetBuilder._write_metadata_local(metadata, metadata_path)
        # Save data processor object.
        self._write_data_processor_local(processor_path)
        return publication_path

    def _publish_s3(self,
                    publication_path: str,
                    copy_path: str,
                    link_path: str,
                    metadata_path: str,
                    processor_path: str,
                    dataset_copy_strategy: str,
                    metadata: dict) -> str:
        """Saves the versioned dataset files to the given S3 path. See publish()
        for more detailed information.

        :param publication_path: The path to which to publish the dataset.
        :param copy_path: The path to which the raw dataset is copied.
        :param link_path: The path to the file containing the raw dataset link.
        :param metadata_path: The path to which the metadata should be saved.
        :param processor_path: The path to which to write the data processor.
        :param dataset_copy_strategy: The strategy by which to copy the
            original, raw dataset to the published path.
        :param metadata: Dataset metadata.
        :return: The versioned dataset's publication path.
        """
        # pylint: disable=too-many-arguments
        fs = S3FileSystem()
        files_to_hash = set()
        # Create publication path.
        VersionedDatasetBuilder._make_publication_path_s3(publication_path, fs)
        # Save tensors.
        self._write_tensors_s3(publication_path, fs)
        # Save the raw dataset.
        if dataset_copy_strategy == STRATEGY_LINK:
            self._make_raw_dataset_link_s3(link_path, fs)
            files_to_hash.add(link_path)
        else:
            file_paths = self._copy_raw_dataset_s3(copy_path, fs)
            files_to_hash = files_to_hash.union(file_paths)
        # Save metadata.
        hash_digest = get_hash_s3(files_to_hash)
        metadata['hash'] = hash_digest
        VersionedDatasetBuilder._write_metadata_s3(metadata, metadata_path, fs)
        # Save data processor object.
        self._write_data_processor_s3(processor_path, fs)
        return publication_path

    @staticmethod
    def _make_publication_path_local(publication_path: str) -> None:
        """Creates the directories that compose the publication path.

        :param publication_path: The path to which to publish the dataset.
        """
        path_obj = Path(publication_path)
        try:
            path_obj.mkdir(parents=True, exist_ok=False)
        except FileExistsError as err:
            raise PublicationPathAlreadyExistsError from err

    @staticmethod
    def _make_publication_path_s3(publication_path: str,
                                  fs: S3FileSystem) -> None:
        """Creates the directories that compose the publication path.

        :param publication_path: The path to which to publish the dataset.
        :param fs: The S3 filesystem object to interface with S3.
        """
        # fs.mkdirs with exist_ok=False does not raise an error, so use ls.
        if fs.ls(publication_path):
            raise PublicationPathAlreadyExistsError
        fs.mkdirs(publication_path)

    def _write_tensors_local(self, publication_path: str) -> set[str]:
        """Writes the feature and label tensors to the publication path
        directory and returns the paths to the created files for hashing.

        :param publication_path: The path to which to publish the dataset.
        :return: The paths to all created files.
        """
        file_paths = set()
        for name, tensor in {**self.features, **self.labels}.items():
            file_path = os.path.join(publication_path, f'{name}.npy')
            file_paths.add(file_path)
            VersionedDatasetBuilder._write_tensor_local(tensor, file_path)
        return file_paths

    def _write_tensors_s3(self,
                          publication_path: str,
                          fs: S3FileSystem) -> set[str]:
        """Writes the feature and label tensors to the publication path
        directory and returns the paths to the created files for hashing.

        :param publication_path: The path to which to publish the dataset.
        :param fs: The S3 filesystem object to interface with S3.
        :return: The paths to all created files.
        """
        file_paths = set()
        for name, tensor in {**self.features, **self.labels}.items():
            file_path = os.path.join(publication_path, f'{name}.npy')
            file_paths.add(file_path)
            VersionedDatasetBuilder._write_tensor_s3(tensor, file_path, fs)
        return file_paths

    @staticmethod
    def _write_tensor_local(tensor: np.ndarray, path: str) -> None:
        """Writes the tensor to the given path.

        :param tensor: The tensor to save.
        :param path: The path to which to save the tensor.
        """
        np.save(path, tensor)

    @staticmethod
    def _write_tensor_s3(tensor: np.ndarray,
                         path: str,
                         fs: S3FileSystem) -> None:
        """Writes the tensor to the given path.

        :param tensor: The tensor to save.
        :param path: The path to which to save the tensor.
        :param fs: The S3 filesystem object to interface with S3.
        """
        with TemporaryFile() as tmp_file:
            np.save(tmp_file, tensor)
            tmp_file.seek(0)
            with fs.open(path, 'wb') as outfile:
                outfile.write(tmp_file.read())

    def _copy_raw_dataset_local(self, copy_path: str) -> set[str]:
        """Copies the raw dataset to the given path, and returns the paths to
        all created files for hashing.

        :param copy_path: The path to which to copy the raw dataset.
        :return: The paths to all created files.
        """
        file_paths = set()
        if self.dataset_path.startswith('s3://'):
            # Copy raw dataset from S3 to local filesystem.
            fs = S3FileSystem()
            fs.get(self.dataset_path, copy_path, recursive=True)
        else:
            # Copy raw dataset from local filesystem to local filesystem.
            shutil.copytree(self.dataset_path, copy_path)
        for current_path, _, filenames in os.walk(copy_path):
            for filename in filenames:
                file_paths.add(os.path.join(current_path, filename))
        return file_paths

    def _copy_raw_dataset_s3(self,
                             copy_path: str,
                             fs: S3FileSystem) -> set[str]:
        """Copies the raw dataset to the given path, and returns the paths to
        all created files for hashing.

        :param copy_path: The path to which to copy the raw dataset.
        :param fs: The S3 filesystem object to interface with S3.
        :return: The paths to all created files.
        """
        s3_file_paths = set()
        if self.dataset_path.startswith('s3://'):
            # Copy raw dataset from S3 to S3.
            dataset_path_no_prefix = self.dataset_path.replace('s3://', '', 1)
            copy_path_no_prefix = copy_path.replace('s3://', '', 1)
            for current_path, _, filenames in fs.walk(self.dataset_path):
                outfile_prefix = current_path.replace(dataset_path_no_prefix,
                                                      copy_path_no_prefix, 1)
                for filename in filenames:
                    infile_path = os.path.join(current_path,
                                               filename)
                    outfile_path = os.path.join(outfile_prefix,
                                                filename)
                    fs.copy(infile_path, outfile_path)
        else:
            # Copy raw dataset from local filesystem to S3.
            fs.put(self.dataset_path, copy_path, recursive=True)
        for current_path, _, filenames in fs.walk(copy_path):
            for filename in filenames:
                s3_file_path = os.path.join(current_path, filename)
                s3_file_paths.add(s3_file_path)
        return s3_file_paths

    def _make_raw_dataset_link_local(self, link_path: str) -> None:
        """Creates a file that contains the path/link to the raw dataset.

        :param link_path: The path to which to create the link file.
        """
        os.mkdir(os.path.dirname(link_path))
        with open(link_path, 'w', encoding='utf-8') as outfile:
            outfile.write(self.dataset_path)

    def _make_raw_dataset_link_s3(self,
                                  link_path: str,
                                  fs: S3FileSystem) -> None:
        """Creates a file that contains the path/link to the raw dataset.

        :param link_path: The path to which to create the link file.
        :param fs: The S3 filesystem object to interface with S3.
        """
        with fs.open(link_path, 'w', encoding='utf-8') as outfile:
            outfile.write(self.dataset_path)

    @staticmethod
    def _write_metadata_local(metadata: dict, metadata_path: str) -> None:
        """Writes the metadata dictionary as a JSON file at the given path.

        :param metadata: The metadata to write.
        :param metadata_path: The path to which to write the metadata.
        """
        with open(metadata_path, 'w', encoding='utf-8') as outfile:
            outfile.write(json.dumps(metadata))

    @staticmethod
    def _write_metadata_s3(metadata: dict,
                           metadata_path: str,
                           fs: S3FileSystem) -> None:
        """Writes the metadata dictionary as a JSON file at the given path.

        :param metadata: The metadata to write.
        :param metadata_path: The path to which to write the metadata.
        :param fs: The S3 filesystem object to interface with S3.
        """
        with fs.open(metadata_path, 'w', encoding='utf-8') as outfile:
            outfile.write(json.dumps(metadata))

    def _write_data_processor_local(self, processor_path: str) -> None:
        """Writes the data processor object to the given path

        :param processor_path: The path to which to write the data processor.
        """
        with open(processor_path, 'wb') as outfile:
            outfile.write(self._serialize_data_processor())

    def _write_data_processor_s3(self,
                                 processor_path: str,
                                 fs: S3FileSystem) -> None:
        """Writes the data processor object to the given path

        :param processor_path: The path to which to write the data processor.
        :param fs: The S3 filesystem object to interface with S3.
        """
        with fs.open(processor_path, 'wb') as outfile:
            outfile.write(self._serialize_data_processor())

    def _serialize_data_processor(self) -> bytes:
        """Returns the serialized representation of the data processor object.

        :return: The serialized representation of the data processor object.
        """
        # See below dill issue on by-value serialization of classes not in
        # __main__:
        # https://github.com/uqfoundation/dill/issues/424
        # The current workaround is provided here:
        # https://github.com/pulumi/pulumi/pull/7755
        obj_type = type(self.data_processor)
        obj_module = obj_type.__module__
        try:
            obj_type.__module__ = '__main__'
            return pickle.dumps(self.data_processor, recurse=True)
        finally:
            obj_type.__module__ = obj_module



"""Tests hashing.py."""

import os
import pytest
from mlops.hashing import hashing
from tests.dataset.test_versioned_dataset_builder import \
    _remove_test_directories_local, _create_test_dataset_local, \
    _remove_test_directories_s3, _create_test_dataset_s3, \
    TEST_DATASET_PATH_LOCAL, TEST_DATASET_FILENAMES, \
    TEST_DATASET_PATH_S3


def test_hash_local_is_reproducible() -> None:
    """Tests that hashing of local files is reproducible."""
    _remove_test_directories_local()
    _create_test_dataset_local()
    files_to_hash_forward = [os.path.join(TEST_DATASET_PATH_LOCAL, filename)
                             for filename in TEST_DATASET_FILENAMES]
    files_to_hash_reverse = [os.path.join(TEST_DATASET_PATH_LOCAL, filename)
                             for filename in TEST_DATASET_FILENAMES[::-1]]
    hash_forward = hashing.get_hash_local(files_to_hash_forward)
    hash_reverse = hashing.get_hash_local(files_to_hash_reverse)
    assert hash_forward
    assert hash_reverse
    assert hash_forward == hash_reverse


@pytest.mark.awstest
def test_hash_s3_is_reproducible() -> None:
    """Tests that hashing of remote files is reproducible."""
    _remove_test_directories_s3()
    _create_test_dataset_s3()
    files_to_hash_forward = [os.path.join(TEST_DATASET_PATH_S3, filename)
                             for filename in TEST_DATASET_FILENAMES]
    files_to_hash_reverse = [os.path.join(TEST_DATASET_PATH_S3, filename)
                             for filename in TEST_DATASET_FILENAMES[::-1]]
    hash_forward = hashing.get_hash_s3(files_to_hash_forward)
    hash_reverse = hashing.get_hash_s3(files_to_hash_reverse)
    assert hash_forward
    assert hash_reverse
    assert hash_forward == hash_reverse


def test_hash_different_if_file_content_different() -> None:
    """Tests that the output of hashing is different if the content of files is
    different."""
    _remove_test_directories_local()
    _create_test_dataset_local()
    files_to_hash = [os.path.join(TEST_DATASET_PATH_LOCAL, filename)
                     for filename in TEST_DATASET_FILENAMES]
    hash_before = hashing.get_hash_local(files_to_hash)
    with open(os.path.join(TEST_DATASET_PATH_LOCAL, TEST_DATASET_FILENAMES[0]),
              'w',
              encoding='utf-8') as outfile:
        outfile.write('new data')
    hash_after = hashing.get_hash_local(files_to_hash)
    assert hash_before != hash_after



"""Publishes a new dataset to the local or remote filesystem. This script should
be run any time the data processor changes."""

import os
from mlops.dataset.versioned_dataset_builder import VersionedDatasetBuilder
from mlops.examples.image.classification.pokemon_classification_data_processor \
    import PokemonClassificationDataProcessor, \
    DEFAULT_DATASET_TRAINVALTEST_PATH

DATASET_VERSION = 'v1'
DATASET_PUBLICATION_PATH_LOCAL = os.path.join('datasets', 'pokemon')
DATASET_PUBLICATION_PATH_S3 = 's3://kosta-mlops/datasets/pokemon'
TAGS = ['image', 'classification']


def publish_dataset(publication_path: str) -> str:
    """Builds and publishes the dataset.

    :param publication_path: The path on the local or remote filesystem to which
        to publish the dataset.
    :return: The versioned dataset's publication path.
    """
    processor = PokemonClassificationDataProcessor()
    builder = VersionedDatasetBuilder(DEFAULT_DATASET_TRAINVALTEST_PATH,
                                      processor)
    return builder.publish(publication_path, DATASET_VERSION, tags=TAGS)


def main() -> None:
    """Runs the program."""
    publish_dataset(DATASET_PUBLICATION_PATH_LOCAL)


if __name__ == '__main__':
    main()



"""Contains custom errors for the Pokemon classification example."""


class LabelsNotFoundError(FileNotFoundError):
    """Raised when a PokemonClassificationDataProcessor attempts to load labels
    for prediction data, an unlabeled data source."""


class NoModelPathsSuppliedError(ValueError):
    """Raised when a non-empty collection of strings representing paths to
    models is expected, but an empty collection is passed instead."""



"""Tests DataProcessor object serialization during dataset publication."""

import os
import shutil
import numpy as np
from mlops.dataset.data_processor import DataProcessor
from mlops.dataset.versioned_dataset_builder import VersionedDatasetBuilder
from mlops.dataset.versioned_dataset import VersionedDataset

TEST_DATASET_PATH = 'sample_data/pokemon/trainvaltest'
TEST_PUBLICATION_PATH = '/tmp/test_serialization/datasets'


class DataProcessorThatWillChange(DataProcessor):
    """A DataProcessor subclass that will change, simulating a user redefining
    how data should enter the model pipeline."""

    def get_raw_features_and_labels(self, dataset_path: str) -> \
            (dict[str, np.ndarray], dict[str, np.ndarray]):
        """Returns dummy features and labels

        :param dataset_path: Unused.
        :return: A 2-tuple of the features dictionary and labels dictionary,
            with matching keys and ordered tensors.
        """
        return ({'X': np.array([1, 2, 3])},
                {'y': np.array([1, 2, 3])})

    def get_raw_features(self, dataset_path: str) -> dict[str, np.ndarray]:
        """Returns dummy features.

        :param dataset_path: Unused.
        :return: A dictionary whose values are feature tensors and whose
            corresponding keys are the names by which those tensors should be
            referenced. For example, the training features (value) may be called
            'X_train' (key).
        """
        return {'X': np.array([1, 2, 3])}

    def preprocess_features(self, raw_feature_tensor: np.ndarray) -> np.ndarray:
        """Returns features multiplied by 2.

        :param raw_feature_tensor: The raw features to be preprocessed.
        :return: The preprocessed feature tensor. This tensor is ready for
            downstream model consumption.
        """
        return 2 * raw_feature_tensor

    def preprocess_labels(self, raw_label_tensor: np.ndarray) -> np.ndarray:
        """Returns labels multiplied by 2.

        :param raw_label_tensor: The raw labels to be preprocessed.
        :return: The preprocessed label tensor. This tensor is ready for
            downstream model consumption.
        """
        return 2 * raw_label_tensor


def _redefine_class() -> None:
    """Redefines DataProcessorThatWillChange."""
    # pylint: disable=redefined-outer-name
    # pylint: disable=global-variable-not-assigned
    # pylint: disable=invalid-name
    # pylint: disable=unused-variable
    global DataProcessorThatWillChange

    class DataProcessorThatWillChange(DataProcessor):
        """A DataProcessor subclass that will change, simulating a user
        redefining how data should enter the model pipeline."""

        def get_raw_features_and_labels(self, dataset_path: str) -> \
                (dict[str, np.ndarray], dict[str, np.ndarray]):
            """Returns dummy features and labels

            :param dataset_path: Unused.
            :return: A 2-tuple of the features dictionary and labels dictionary,
                with matching keys and ordered tensors.
            """
            raise ValueError('The new implementation is different.')

        def get_raw_features(self, dataset_path: str) -> dict[str, np.ndarray]:
            """Returns dummy features.

            :param dataset_path: Unused.
            :return: A dictionary whose values are feature tensors and whose
                corresponding keys are the names by which those tensors should
                be referenced. For example, the training features (value) may be
                called 'X_train' (key).
            """
            raise ValueError('The new implementation is different.')

        def preprocess_features(self, raw_feature_tensor: np.ndarray) -> \
                np.ndarray:
            """Returns features multiplied by 2.

            :param raw_feature_tensor: The raw features to be preprocessed.
            :return: The preprocessed feature tensor. This tensor is ready for
                downstream model consumption.
            """
            raise ValueError('The new implementation is different.')

        def preprocess_labels(self, raw_label_tensor: np.ndarray) -> np.ndarray:
            """Returns labels multiplied by 2.

            :param raw_label_tensor: The raw labels to be preprocessed.
            :return: The preprocessed label tensor. This tensor is ready for
                downstream model consumption.
            """
            raise ValueError('The new implementation is different.')


def test_serialized_data_processor_uses_original_methods() -> None:
    """Tests that the serialized data processor object obtained during dataset
    publication uses the original methods, not those of the class definition at
    the time of deserialization, which may be different."""
    try:
        shutil.rmtree(TEST_PUBLICATION_PATH)
    except FileNotFoundError:
        pass
    data_processor = DataProcessorThatWillChange()
    builder = VersionedDatasetBuilder(TEST_DATASET_PATH, data_processor)
    builder.publish(TEST_PUBLICATION_PATH, 'v1')
    dataset = VersionedDataset(os.path.join(TEST_PUBLICATION_PATH, 'v1'))
    features, labels = dataset.data_processor.get_raw_features_and_labels('dne')
    assert np.array_equal(features['X'], [1, 2, 3])
    assert np.array_equal(labels['y'], [1, 2, 3])
    _redefine_class()
    dataset = VersionedDataset(os.path.join(TEST_PUBLICATION_PATH, 'v1'))
    features, labels = dataset.data_processor.get_raw_features_and_labels('dne')
    assert np.array_equal(features['X'], [1, 2, 3])
    assert np.array_equal(labels['y'], [1, 2, 3])



"""Tests versioned_dataset.py."""

import os
import pytest
from mlops.dataset.versioned_dataset_builder import VersionedDatasetBuilder
from mlops.dataset.versioned_dataset import VersionedDataset
from tests.dataset.test_versioned_dataset_builder import \
    _remove_test_directories_local, _create_test_dataset_local, \
    _remove_test_directories_s3, \
    TEST_DATASET_PATH_LOCAL, TEST_PUBLICATION_PATH_LOCAL, \
    TEST_PUBLICATION_PATH_S3
from tests.dataset.preset_data_processor import PresetDataProcessor

EXPECTED_ATTRIBUTES = {'X_train', 'X_val', 'X_test',
                       'y_train', 'y_val', 'y_test',
                       'md5',
                       'data_processor'}


def _publish_test_dataset_local() -> None:
    """Publishes the versioned dataset to the local filesystem."""
    _remove_test_directories_local()
    _create_test_dataset_local()
    processor = PresetDataProcessor()
    builder = VersionedDatasetBuilder(TEST_DATASET_PATH_LOCAL, processor)
    version = 'v1'
    builder.publish(TEST_PUBLICATION_PATH_LOCAL, version)


def _publish_test_dataset_s3() -> None:
    """Publishes the versioned dataset to S3."""
    _remove_test_directories_local()
    _create_test_dataset_local()
    _remove_test_directories_s3()
    processor = PresetDataProcessor()
    builder = VersionedDatasetBuilder(TEST_DATASET_PATH_LOCAL, processor)
    version = 'v1'
    builder.publish(TEST_PUBLICATION_PATH_S3, version)


def test_init_creates_expected_attributes_local() -> None:
    """Tests that init creates all feature/label tensors, the hash, and the
    data processor as attributes when the dataset is on the local filesystem."""
    _publish_test_dataset_local()
    dataset = VersionedDataset(os.path.join(TEST_PUBLICATION_PATH_LOCAL, 'v1'))
    for attribute in EXPECTED_ATTRIBUTES:
        assert hasattr(dataset, attribute)


@pytest.mark.awstest
def test_init_creates_expected_attributes_s3() -> None:
    """Tests that init creates all feature/label tensors, the hash, and the
    data processor as attributes when the dataset is on S3."""
    _publish_test_dataset_s3()
    dataset = VersionedDataset(os.path.join(TEST_PUBLICATION_PATH_S3, 'v1'))
    for attribute in EXPECTED_ATTRIBUTES:
        assert hasattr(dataset, attribute)


def test_versioned_datasets_from_same_files_are_equal() -> None:
    """Tests that two versioned datasets loaded from the same files are
    considered equal in comparisons."""
    _publish_test_dataset_local()
    dataset_path = os.path.join(TEST_PUBLICATION_PATH_LOCAL, 'v1')
    dataset1 = VersionedDataset(dataset_path)
    dataset2 = VersionedDataset(dataset_path)
    assert dataset1 == dataset2


def test_hashcode_is_hash_of_md5_digest() -> None:
    """Tests that the hashcode of the dataset object is the hash of the loaded
    MD5 digest."""
    _publish_test_dataset_local()
    dataset = VersionedDataset(os.path.join(TEST_PUBLICATION_PATH_LOCAL, 'v1'))
    assert hash(dataset) == hash(dataset.md5)



"""Tests train_model.py."""
# pylint: disable=no-name-in-module,no-member

import os
import shutil
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Flatten
from mlops.dataset.versioned_dataset import VersionedDataset
from mlops.examples.image.classification.publish_dataset import \
    publish_dataset, DATASET_VERSION
from mlops.examples.image.classification import train_model

TEST_DATASET_PUBLICATION_PATH_LOCAL = '/tmp/test_train_model/datasets'
TEST_CHECKPOINT_PATH = '/tmp/test_train_model/models/checkpoints'
TEST_MODEL_PUBLICATION_PATH_LOCAL = '/tmp/test_train_model/models/versioned'


def _create_dataset() -> None:
    """Creates the dataset files."""
    try:
        shutil.rmtree(TEST_DATASET_PUBLICATION_PATH_LOCAL)
    except FileNotFoundError:
        pass
    publish_dataset(TEST_DATASET_PUBLICATION_PATH_LOCAL)


def test_get_baseline_model_correct_shapes() -> None:
    """Tests that the baseline model has the correct input and output shapes."""
    _create_dataset()
    dataset = VersionedDataset(os.path.join(TEST_DATASET_PUBLICATION_PATH_LOCAL,
                                            DATASET_VERSION))
    model = train_model.get_baseline_model(dataset)
    assert model.input_shape[1:] == dataset.X_train.shape[1:]
    assert model.output_shape[1:] == dataset.y_train.shape[1:]


def test_train_model_creates_checkpoints() -> None:
    """Tests that train_model creates model checkpoints."""
    _create_dataset()
    try:
        shutil.rmtree(TEST_CHECKPOINT_PATH)
    except FileNotFoundError:
        pass
    dataset = VersionedDataset(os.path.join(TEST_DATASET_PUBLICATION_PATH_LOCAL,
                                            DATASET_VERSION))
    model = Sequential([
        Flatten(input_shape=dataset.X_train.shape[1:]),
        Dense(dataset.y_train.shape[1])])
    model.compile('adam', loss='mse')
    train_kwargs = {'epochs': 3, 'batch_size': 8}
    model_checkpoint_filename = os.path.join(TEST_CHECKPOINT_PATH, 'model.h5')
    _ = train_model.train_model(
        model,
        dataset,
        model_checkpoint_filename=model_checkpoint_filename,
        **train_kwargs)
    assert os.path.exists(model_checkpoint_filename)


def test_train_model_returns_correct_training_config() -> None:
    """Tests that train_model returns a TrainingConfig object with the correct
    information."""
    _create_dataset()
    dataset = VersionedDataset(os.path.join(TEST_DATASET_PUBLICATION_PATH_LOCAL,
                                            DATASET_VERSION))
    model = Sequential([
        Flatten(input_shape=dataset.X_train.shape[1:]),
        Dense(dataset.y_train.shape[1])])
    model.compile('adam', loss='mse')
    train_kwargs = {'epochs': 3, 'batch_size': 8}
    model_checkpoint_filename = os.path.join(TEST_CHECKPOINT_PATH, 'model.h5')
    training_config = train_model.train_model(
        model,
        dataset,
        model_checkpoint_filename=model_checkpoint_filename,
        **train_kwargs)
    assert training_config.history == model.history
    assert training_config.train_args == train_kwargs


def test_publish_model_creates_files() -> None:
    """Tests that publish_model creates the published model files."""
    _create_dataset()
    try:
        shutil.rmtree(TEST_MODEL_PUBLICATION_PATH_LOCAL)
    except FileNotFoundError:
        pass
    dataset = VersionedDataset(os.path.join(TEST_DATASET_PUBLICATION_PATH_LOCAL,
                                            DATASET_VERSION))
    model = Sequential([
        Flatten(input_shape=dataset.X_train.shape[1:]),
        Dense(dataset.y_train.shape[1])])
    model.compile('adam', loss='mse')
    train_kwargs = {'epochs': 3, 'batch_size': 8}
    model_checkpoint_filename = os.path.join(TEST_CHECKPOINT_PATH, 'model.h5')
    training_config = train_model.train_model(
        model,
        dataset,
        model_checkpoint_filename=model_checkpoint_filename,
        **train_kwargs)
    train_model.publish_model(
        model,
        dataset,
        training_config,
        TEST_MODEL_PUBLICATION_PATH_LOCAL)
    assert os.path.exists(TEST_MODEL_PUBLICATION_PATH_LOCAL)
    assert len(os.listdir(TEST_MODEL_PUBLICATION_PATH_LOCAL)) == 1



"""Contains the VersionedModelBuilder class."""
# pylint: disable=no-name-in-module

import os
import json
from tempfile import TemporaryDirectory
from typing import Optional
from datetime import datetime
from s3fs import S3FileSystem
from tensorflow.keras.models import Model
from mlops.dataset.versioned_dataset_builder import VersionedDatasetBuilder
from mlops.dataset.versioned_dataset import VersionedDataset
from mlops.model.training_config import TrainingConfig
from mlops.hashing.hashing import get_hash_local, get_hash_s3


class VersionedModelBuilder:
    """An object containing all of the components that form a versioned
    model."""
    # pylint: disable=too-few-public-methods

    def __init__(self,
                 versioned_dataset: VersionedDataset,
                 model: Model,
                 training_config: TrainingConfig) -> None:
        """Instantiates the object.

        :param versioned_dataset: The versioned dataset object with which the
            model was trained/validated/tested. Used to preprocess new data at
            prediction time.
        :param model: The trained Keras model. This is the exact version of the
            model that will be saved; if you intend to keep the best weights and
            not the ones with which the model finished training, ensure that you
            set the model's weights to those desired (this can be done easily
            using the ModelCheckpoint callback or a custom callback that stores
            the best weights in memory).
        :param training_config: The model's training configuration.
        """
        self.versioned_dataset = versioned_dataset
        self.model = model
        self.training_config = training_config

    def publish(self,
                path: str,
                version: Optional[str] = None,
                tags: Optional[list[str]] = None) -> str:
        """Saves the versioned model files to the given path. If the path and
        appended version already exists, this operation will raise a
        PublicationPathAlreadyExistsError.

        The following files will be created:
            path/version/ (the publication path and version)
                model.h5 (the saved model)
                meta.json (metadata)

        The contents of meta.json will be:
            {
                version: (model version)
                hash: (MD5 hash of all objects apart from meta.json)
                dataset: (the link to the dataset used during training)
                history: (the training history dictionary)
                train_args: (the training arguments dictionary)
                created_at: (timestamp)
                tags: (optional list of tags)
            }

        :param path: The path, either on the local filesystem or in a cloud
            store such as S3, to which the model should be saved. The version
            will be appended to this path as a subdirectory. An S3 path
            should be a URL of the form "s3://bucket-name/path/to/dir". It is
            recommended to use this same path to publish all models, since it
            will prevent the user from creating two different models with the
            same version.
        :param version: A string indicating the model version. The version
            should be unique to this model. If None, the publication timestamp
            will be used as the version.
        :param tags: An optional list of string tags to add to the model
            metadata.
        :return: The versioned model's publication path.
        """
        timestamp = datetime.now().isoformat()
        if not version:
            version = timestamp
        if not tags:
            tags = []
        publication_path = os.path.join(path.rstrip('/'), version)
        model_path = os.path.join(publication_path, 'model.h5')
        metadata_path = os.path.join(publication_path, 'meta.json')
        metadata = {
            'version': version,
            'hash': 'TDB',
            'dataset': self.versioned_dataset.path,
            'history': self.training_config.history.history,
            'train_args': self.training_config.train_args,
            'created_at': timestamp,
            'tags': tags}
        if path.startswith('s3://'):
            return self._publish_s3(publication_path,
                                    model_path,
                                    metadata_path,
                                    metadata)
        return self._publish_local(publication_path,
                                   model_path,
                                   metadata_path,
                                   metadata)

    def _publish_local(self,
                       publication_path: str,
                       model_path: str,
                       metadata_path: str,
                       metadata: dict) -> str:
        """Saves the versioned model files to the given local path. See
        publish() for more detailed information.

        :param publication_path: The local path to which to publish the model.
        :param model_path: The path to which the model is saved.
        :param metadata_path: The path to which the metadata is saved.
        :param metadata: Model metadata.
        :return: The versioned model's publication path.
        """
        files_to_hash = set()
        # Create publication path.
        # pylint: disable=protected-access
        VersionedDatasetBuilder._make_publication_path_local(publication_path)
        # Save model.
        self.model.save(model_path)
        files_to_hash.add(model_path)
        # Save metadata.
        hash_digest = get_hash_local(files_to_hash)
        metadata['hash'] = hash_digest
        with open(metadata_path, 'w', encoding='utf-8') as outfile:
            outfile.write(json.dumps(metadata))
        return publication_path

    def _publish_s3(self,
                    publication_path: str,
                    model_path: str,
                    metadata_path: str,
                    metadata: dict) -> str:
        """Saves the versioned model files to the given S3 path. See publish()
        for more detailed information.

        :param publication_path: The S3 path to which to publish the model.
        :param model_path: The path to which the model is saved.
        :param metadata_path: The path to which the metadata is saved.
        :param metadata: Model metadata.
        :return: The versioned model's publication path.
        """
        fs = S3FileSystem()
        files_to_hash = set()
        # Create publication path.
        # pylint: disable=protected-access
        VersionedDatasetBuilder._make_publication_path_s3(publication_path, fs)
        # Save model.
        with TemporaryDirectory() as tmp_dir:
            self.model.save(f'{tmp_dir}/model.h5')
            fs.put(f'{tmp_dir}/model.h5', model_path)
        files_to_hash.add(model_path)
        # Save metadata.
        hash_digest = get_hash_s3(files_to_hash)
        metadata['hash'] = hash_digest
        with fs.open(metadata_path, 'w', encoding='utf-8') as outfile:
            outfile.write(json.dumps(metadata))
        return publication_path



"""Tests connection to S3."""

import os
import pytest
import boto3
from botocore.exceptions import ClientError

PROJECT_S3_BUCKET = 'kosta-mlops'
TEST_FILENAME = 's3_conn_test.txt'
TEST_FILE_PATH = os.path.join('/tmp', TEST_FILENAME)


@pytest.mark.awstest
def test_connection_to_project_s3_bucket() -> None:
    """Tests that the project S3 bucket (s3://kosta-mlops) can be reached."""
    s3 = boto3.client('s3')
    response = s3.list_buckets()
    bucket_matches = [bucket_dict for bucket_dict in response['Buckets']
                      if bucket_dict['Name'] == PROJECT_S3_BUCKET]
    assert len(bucket_matches) == 1


@pytest.mark.awstest
def test_project_s3_bucket_read_write() -> None:
    """Tests that the project S3 bucket (s3://kosta-mlops) can be read from and
    written to."""
    s3 = boto3.client('s3')
    # Remove test file from local filesystem.
    try:
        os.remove(TEST_FILE_PATH)
    except FileNotFoundError:
        pass
    # Remove test file from S3.
    s3.delete_object(Bucket=PROJECT_S3_BUCKET, Key=TEST_FILENAME)
    # Create test file on local filesystem.
    with open(TEST_FILE_PATH, 'w', encoding='utf-8') as outfile:
        outfile.write(TEST_FILENAME)
    # Attempt to upload to S3.
    try:
        _ = s3.upload_file(TEST_FILE_PATH, PROJECT_S3_BUCKET, TEST_FILENAME)
    except ClientError as exc:
        raise exc
    # Remove test file from local filesystem.
    os.remove(TEST_FILE_PATH)
    # Read test file from S3.
    s3.download_file(PROJECT_S3_BUCKET, TEST_FILENAME, TEST_FILE_PATH)
    # Remove test file from S3.
    s3.delete_object(Bucket=PROJECT_S3_BUCKET, Key=TEST_FILENAME)
    # Check that contents of downloaded file match what we sent for upload.
    with open(TEST_FILE_PATH, 'r', encoding='utf-8') as infile:
        assert infile.read() == TEST_FILENAME



"""Tests errors.py."""

from mlops.examples.image.classification.errors import LabelsNotFoundError, \
    NoModelPathsSuppliedError


def test_labels_not_found_error_extends_file_not_found_error() -> None:
    """Tests that LabelsNotFoundError extends FileNotFoundError."""
    err = LabelsNotFoundError()
    assert isinstance(err, FileNotFoundError)


def test_no_model_paths_supplied_error_extends_value_error() -> None:
    """Tests that NoModelPathsSuppliedError extends ValueError."""
    err = NoModelPathsSuppliedError()
    assert isinstance(err, ValueError)



"""Tests data_processor.py."""

import numpy as np
import pytest
from mlops.dataset.data_processor import DataProcessor
from tests.dataset.preset_data_processor import PresetDataProcessor


@pytest.mark.xfail
def test_data_processor_is_abstract() -> None:
    """Tests that the DataProcessor object is abstract."""
    # This test currently fails because we needed to make DataProcessor concrete
    # as a serialization workaround. See:
    # https://github.com/kostaleonard/mlops/issues/12
    with pytest.raises(TypeError):
        _ = DataProcessor()


def test_data_processor_accepts_local_path() -> None:
    """Tests that the DataProcessor object accepts paths on the local
    filesystem."""
    processor = PresetDataProcessor()
    features, labels = processor.get_preprocessed_features_and_labels(
        'path/to/dataset')
    assert set(features.keys()) == {'X_train', 'X_val', 'X_test'}
    assert set(labels.keys()) == {'y_train', 'y_val', 'y_test'}


def test_data_processor_accepts_remote_path() -> None:
    """Tests that the DataProcessor object accepts paths on remote filesystens,
    e.g., S3."""
    processor = PresetDataProcessor()
    features, labels = processor.get_preprocessed_features_and_labels(
        's3://path/to/my/dataset')
    assert set(features.keys()) == {'X_train', 'X_val', 'X_test'}
    assert set(labels.keys()) == {'y_train', 'y_val', 'y_test'}


def test_raw_feature_keys_match_preprocessed_feature_keys() -> None:
    """Tests that the dictionaries returned by get_preprocessed_features and
    get_raw_features have the same keys."""
    processor = PresetDataProcessor()
    features = processor.get_preprocessed_features('path/to/my/features')
    raw_features = processor.get_raw_features('path/to/my/features')
    assert set(features.keys()) == set(raw_features.keys())


def test_raw_label_keys_match_preprocessed_label_keys() -> None:
    """Tests that the dictionaries returned by get_preprocessed_labels and
    get_raw_labels have the same keys."""
    processor = PresetDataProcessor()
    _, labels = processor.get_preprocessed_features_and_labels(
        'path/to/my/dataset')
    _, raw_labels = processor.get_raw_features_and_labels(
        'path/to/my/dataset')
    assert set(labels.keys()) == set(raw_labels.keys())


def test_get_raw_features_returns_raw_features() -> None:
    """Tests that get_raw_features returns the raw features, before
    preprocessing is applied."""
    processor = PresetDataProcessor()
    features = processor.get_preprocessed_features('path/to/my/features')
    raw_features = processor.get_raw_features('path/to/my/features')
    for name in features.keys():
        assert not np.array_equal(raw_features[name], features[name])


def test_get_raw_labels_returns_raw_labels() -> None:
    """Tests that get_raw_labels returns the raw labels, before
    preprocessing is applied."""
    processor = PresetDataProcessor()
    _, labels = processor.get_preprocessed_features_and_labels(
        'path/to/my/dataset')
    _, raw_labels = processor.get_raw_features_and_labels('path/to/my/dataset')
    for name in labels.keys():
        assert not np.array_equal(raw_labels[name], labels[name])


def test_preprocessed_features_match() -> None:
    """Tests that the result of applying preprocess_features on the raw features
    is the same as the output of get_preprocessed_features."""
    processor = PresetDataProcessor()
    features = processor.get_preprocessed_features('path/to/my/features')
    raw_features = processor.get_raw_features('path/to/my/features')
    for name, raw in raw_features.items():
        manually_preprocessed_features = processor.preprocess_features(raw)
        assert np.array_equal(manually_preprocessed_features, features[name])


def test_preprocessed_labels_match() -> None:
    """Tests that the result of applying preprocess_labels on the raw labels
    is the same as the output of get_preprocessed_labels."""
    processor = PresetDataProcessor()
    _, labels = processor.get_preprocessed_features_and_labels(
        'path/to/my/dataset')
    _, raw_labels = processor.get_raw_features_and_labels('path/to/my/dataset')
    for name, raw in raw_labels.items():
        manually_preprocessed_labels = processor.preprocess_labels(raw)
        assert np.array_equal(manually_preprocessed_labels, labels[name])


def test_get_raw_features_and_labels_features_match() -> None:
    """Tests that get_raw_features_and_labels and get_raw_features have matching
    features."""
    processor = PresetDataProcessor()
    features, _ = processor.get_raw_features_and_labels('path/to/dataset')
    features_only = processor.get_raw_features('path/to/my/features')
    assert set(features.keys()) == set(features_only.keys())
    for name, arr in features.items():
        assert np.array_equal(arr, features_only[name])


def test_get_preprocessed_features_and_labels_features_match() -> None:
    """Tests that get_preprocessed_features_and_labels and
    get_preprocessed_features have matching features."""
    processor = PresetDataProcessor()
    features, _ = processor.get_preprocessed_features_and_labels(
        'path/to/dataset')
    features_only = processor.get_preprocessed_features('path/to/my/features')
    assert set(features.keys()) == set(features_only.keys())
    for name in features:
        assert np.array_equal(features[name], features_only[name])



"""Tests versioned_model_builder.py."""
# pylint: disable=redefined-outer-name,no-name-in-module

import os
import shutil
from datetime import datetime
import json
import numpy as np
import pytest
from s3fs import S3FileSystem
from tensorflow.keras.models import Model, Sequential, load_model
from tensorflow.keras.layers import Dense
from mlops.dataset.versioned_dataset import VersionedDataset
from mlops.model.versioned_model_builder import VersionedModelBuilder
from mlops.model.training_config import TrainingConfig
from mlops.errors import PublicationPathAlreadyExistsError
from tests.dataset.test_versioned_dataset import _publish_test_dataset_local, \
    TEST_PUBLICATION_PATH_LOCAL as TEST_DATASET_PUBLICATION_PATH_LOCAL

TEST_MODEL_PUBLICATION_PATH_LOCAL = '/tmp/test_versioned_model_builder/model'
TEST_MODEL_PUBLICATION_PATH_S3 = ('s3://kosta-mlops/'
                                  'test_versioned_model_builder/model')


def _remove_test_directories_local() -> None:
    """Removes the local test directories."""
    try:
        shutil.rmtree(TEST_MODEL_PUBLICATION_PATH_LOCAL)
    except FileNotFoundError:
        pass


def _remove_test_directories_s3() -> None:
    """Removes the S3 test directories."""
    fs = S3FileSystem()
    try:
        fs.rm(TEST_MODEL_PUBLICATION_PATH_S3, recursive=True)
    except FileNotFoundError:
        pass


@pytest.fixture
def dataset() -> VersionedDataset:
    """Returns the versioned dataset fixture for testing.

    :return: The versioned dataset fixture.
    """
    _publish_test_dataset_local()
    return VersionedDataset(os.path.join(TEST_DATASET_PUBLICATION_PATH_LOCAL,
                                         'v1'))


@pytest.fixture
def model(dataset: VersionedDataset) -> Model:
    """Returns the model fixture for testing.

    :param dataset: The versioned dataset.
    :return: The model fixture.
    """
    mod = Sequential([Dense(dataset.y_train.shape[1],
                            input_shape=dataset.X_train.shape[1:])])
    mod.compile('adam', loss='mse')
    return mod


@pytest.fixture
def training_config(dataset: VersionedDataset,
                    model: Model) -> TrainingConfig:
    """Returns the training configuration fixture for testing.

    :param dataset: The versioned dataset.
    :param model: The model.
    :return: The training configuration fixture.
    """
    train_kwargs = {'epochs': 5,
                    'batch_size': 8}
    history = model.fit(x=dataset.X_train,
                        y=dataset.y_train,
                        **train_kwargs)
    return TrainingConfig(history, train_kwargs)


def test_publish_appends_explicit_version(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that publish appends the version string to the path.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder = VersionedModelBuilder(dataset, model, training_config)
    version = 'v2'
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, version=version)
    expected_filename = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL, version)
    assert os.path.exists(expected_filename)
    assert os.path.isdir(expected_filename)


def test_publish_appends_version_timestamp(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that publish appends the timestamp to the path when no version is
    given.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder = VersionedModelBuilder(dataset, model, training_config)
    start = datetime.now()
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL)
    end = datetime.now()
    assert len(os.listdir(TEST_MODEL_PUBLICATION_PATH_LOCAL)) == 1
    dirname = os.listdir(TEST_MODEL_PUBLICATION_PATH_LOCAL)[0]
    publication_time = datetime.fromisoformat(dirname)
    assert start < publication_time < end


def test_publish_local_path_creates_expected_files(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that publish on a local path creates the expected
    files/directories on the local filesystem.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder = VersionedModelBuilder(dataset, model, training_config)
    version = 'v2'
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, version=version)
    assert len(os.listdir(TEST_MODEL_PUBLICATION_PATH_LOCAL)) == 1
    assert os.listdir(TEST_MODEL_PUBLICATION_PATH_LOCAL)[0] == version
    publication_dir = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL, version)
    assert set(os.listdir(publication_dir)) == {'model.h5', 'meta.json'}


@pytest.mark.awstest
def test_publish_s3_path_creates_expected_files(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that publish on an S3 path creates the expected files/directories
    on S3.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_s3()
    builder = VersionedModelBuilder(dataset, model, training_config)
    version = 'v2'
    builder.publish(TEST_MODEL_PUBLICATION_PATH_S3, version=version)
    fs = S3FileSystem()
    expected_filename = os.path.join(TEST_MODEL_PUBLICATION_PATH_S3, version)
    # Remove 's3://' from latter paths.
    assert set(fs.ls(expected_filename)) == {
        os.path.join(expected_filename, 'model.h5')[5:],
        os.path.join(expected_filename, 'meta.json')[5:]}
    assert fs.isdir(expected_filename)


def test_publish_local_path_raises_path_already_exists_error(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that publish on a local path that already exists raises a
    PublicationPathAlreadyExistsError.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder = VersionedModelBuilder(dataset, model, training_config)
    version = 'v2'
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, version=version)
    with pytest.raises(PublicationPathAlreadyExistsError):
        builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, version=version)


@pytest.mark.awstest
def test_publish_s3_path_raises_path_already_exists_error(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that publish on an S3 path that already exists raises a
    PublicationPathAlreadyExistsError.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_s3()
    builder = VersionedModelBuilder(dataset, model, training_config)
    version = 'v2'
    builder.publish(TEST_MODEL_PUBLICATION_PATH_S3, version=version)
    with pytest.raises(PublicationPathAlreadyExistsError):
        builder.publish(TEST_MODEL_PUBLICATION_PATH_S3, version=version)


def test_publish_includes_expected_metadata(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that publish creates a file meta.json with the expected
    metadata.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder = VersionedModelBuilder(dataset, model, training_config)
    version = 'v2'
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, version=version)
    meta_path = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL, version,
                             'meta.json')
    with open(meta_path, 'r', encoding='utf-8') as infile:
        contents = json.loads(infile.read())
    assert set(contents.keys()) == {
        'version',
        'hash',
        'dataset',
        'history',
        'train_args',
        'created_at',
        'tags'}


def test_publish_timestamps_match(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that all 3 timestamps match if no version string is supplied:
    metadata.json's version and created_at fields, and the final directory
    of the published path.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder = VersionedModelBuilder(dataset, model, training_config)
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL)
    assert len(os.listdir(TEST_MODEL_PUBLICATION_PATH_LOCAL)) == 1
    dirname = os.listdir(TEST_MODEL_PUBLICATION_PATH_LOCAL)[0]
    meta_path = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL, dirname,
                             'meta.json')
    with open(meta_path, 'r', encoding='utf-8') as infile:
        contents = json.loads(infile.read())
    version_time = contents['version']
    created_at_time = contents['created_at']
    assert dirname == version_time == created_at_time


def test_publish_local_with_trailing_slash(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that publishing to a local path with a trailing slash works
    properly.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder = VersionedModelBuilder(dataset, model, training_config)
    version = 'v2'
    # One trailing slash.
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL + '/', version)
    expected_filename = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL, version)
    assert os.path.exists(expected_filename)
    assert os.path.isdir(expected_filename)
    _remove_test_directories_local()
    # Many trailing slashes.
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL + '///', version)
    assert os.path.exists(expected_filename)
    assert os.path.isdir(expected_filename)


@pytest.mark.awstest
def test_publish_s3_with_trailing_slash(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that publishing to an S3 path with a trailing slash works
    properly.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_s3()
    builder = VersionedModelBuilder(dataset, model, training_config)
    version = 'v2'
    # One trailing slash.
    builder.publish(TEST_MODEL_PUBLICATION_PATH_S3 + '/', version=version)
    fs = S3FileSystem()
    expected_filename = os.path.join(TEST_MODEL_PUBLICATION_PATH_S3, version)
    assert fs.exists(expected_filename)
    assert fs.isdir(expected_filename)
    # Many trailing slashes.
    _remove_test_directories_s3()
    builder.publish(TEST_MODEL_PUBLICATION_PATH_S3 + '///', version=version)
    fs = S3FileSystem()
    expected_filename = os.path.join(TEST_MODEL_PUBLICATION_PATH_S3, version)
    assert fs.exists(expected_filename)
    assert fs.isdir(expected_filename)


def test_same_models_have_same_hashes(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that the hash values from two models that have identical files are
    the same.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder = VersionedModelBuilder(dataset, model, training_config)
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1')
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v2')
    meta_path1 = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL,
                              'v1',
                              'meta.json')
    meta_path2 = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL,
                              'v2',
                              'meta.json')
    with open(meta_path1, 'r', encoding='utf-8') as infile:
        contents1 = json.loads(infile.read())
    with open(meta_path2, 'r', encoding='utf-8') as infile:
        contents2 = json.loads(infile.read())
    assert contents1['created_at'] != contents2['created_at']
    assert contents1['hash'] == contents2['hash']


def test_different_models_have_different_hashes(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that the hash values from two models that have different files are
    different.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder1 = VersionedModelBuilder(dataset, model, training_config)
    builder1.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1')
    # model2 has an extra layer.
    model2 = Sequential([Dense(5, input_shape=dataset.X_train.shape[1:]),
                         Dense(dataset.y_train.shape[1])])
    model2.compile('adam', loss='mse')
    train_args = {'epochs': 1, 'batch_size': 2}
    history = model2.fit(x=dataset.X_train,
                         y=dataset.y_train,
                         **train_args)
    training_config2 = TrainingConfig(history, train_args)
    builder2 = VersionedModelBuilder(dataset, model2, training_config2)
    builder2.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v2')
    meta_path1 = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL,
                              'v1',
                              'meta.json')
    meta_path2 = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL,
                              'v2',
                              'meta.json')
    with open(meta_path1, 'r', encoding='utf-8') as infile:
        contents1 = json.loads(infile.read())
    with open(meta_path2, 'r', encoding='utf-8') as infile:
        contents2 = json.loads(infile.read())
    assert contents1['created_at'] != contents2['created_at']
    assert contents1['hash'] != contents2['hash']


@pytest.mark.awstest
def test_publish_local_and_s3_create_same_model(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that publishing locally or remotely on S3 produces the same model.
    Verifies identity by comparing model hashes.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    _remove_test_directories_s3()
    builder = VersionedModelBuilder(dataset, model, training_config)
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1')
    builder.publish(TEST_MODEL_PUBLICATION_PATH_S3, 'v1')
    meta_path1 = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL,
                              'v1',
                              'meta.json')
    meta_path2 = os.path.join(TEST_MODEL_PUBLICATION_PATH_S3,
                              'v1',
                              'meta.json')
    with open(meta_path1, 'r', encoding='utf-8') as infile:
        contents1 = json.loads(infile.read())
    fs = S3FileSystem()
    with fs.open(meta_path2, 'r', encoding='utf-8') as infile:
        contents2 = json.loads(infile.read())
    assert contents1['created_at'] != contents2['created_at']
    assert contents1['hash'] == contents2['hash']


def test_metadata_includes_training_history(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that the metadata includes the training history dictionary, and
    that it is consistent with the training results.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder = VersionedModelBuilder(dataset, model, training_config)
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1')
    meta_path = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1',
                             'meta.json')
    with open(meta_path, 'r', encoding='utf-8') as infile:
        contents = json.loads(infile.read())
    meta_history = contents['history']
    assert meta_history == training_config.history.history


def test_metadata_includes_training_args(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that the metadata includes any kwargs supplied to the training
    function.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder = VersionedModelBuilder(dataset, model, training_config)
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1')
    meta_path = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1',
                             'meta.json')
    with open(meta_path, 'r', encoding='utf-8') as infile:
        contents = json.loads(infile.read())
    meta_args = contents['train_args']
    assert meta_args == training_config.train_args


def test_metadata_includes_dataset_link(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that the metadata includes a link to the dataset.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder = VersionedModelBuilder(dataset, model, training_config)
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1')
    meta_path = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1',
                             'meta.json')
    with open(meta_path, 'r', encoding='utf-8') as infile:
        contents = json.loads(infile.read())
    link = contents['dataset']
    assert link == dataset.path


def test_metadata_includes_expected_tags(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that the metadata includes the expected tags.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder = VersionedModelBuilder(dataset, model, training_config)
    tags = ['hello', 'world']
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1', tags=tags)
    meta_path = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1',
                             'meta.json')
    with open(meta_path, 'r', encoding='utf-8') as infile:
        contents = json.loads(infile.read())
    meta_tags = contents['tags']
    assert meta_tags == tags


def test_published_model_performance_matches_trained_model(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that the published model has the same performance as the model
    supplied during object instantiation.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder = VersionedModelBuilder(dataset, model, training_config)
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1')
    saved_model_path = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL,
                                    'v1',
                                    'model.h5')
    loaded_model = load_model(saved_model_path)
    val_err = model.evaluate(x=dataset.X_val, y=dataset.y_val)
    loaded_val_err = loaded_model.evaluate(x=dataset.X_val, y=dataset.y_val)
    assert np.isclose(val_err, loaded_val_err)



"""Tests __init__.py."""

from mlops.errors import PublicationPathAlreadyExistsError, \
    InvalidDatasetCopyStrategyError


def test_publication_path_error_extends_file_exists_error() -> None:
    """Tests that PublicationPathAlreadyExistsError extends FileExistsError."""
    err = PublicationPathAlreadyExistsError()
    assert isinstance(err, FileExistsError)


def test_invalid_dataset_copy_strategy_error_extends_value_error() -> None:
    """Tests that InvalidDatasetCopyStrategyError extends ValueError."""
    err = InvalidDatasetCopyStrategyError()
    assert isinstance(err, ValueError)



"""Contains the PokemonClassificationDataProcessor class."""

import os
from typing import Optional
import numpy as np
import pandas as pd
from matplotlib.image import imread
from mlops.dataset.invertible_data_processor import InvertibleDataProcessor
from mlops.examples.image.classification.errors import LabelsNotFoundError

DEFAULT_DATASET_TRAINVALTEST_PATH = os.path.join('sample_data', 'pokemon',
                                                 'trainvaltest')
DEFAULT_DATASET_PRED_PATH = os.path.join('sample_data', 'pokemon', 'pred')
IMAGES_DIRNAME = 'images'
LABELS_FILENAME = 'pokemon.csv'
TRAIN_SPLIT = 0.7
VAL_SPLIT = 0.2
CLASSES = ['Normal', 'Fire', 'Water', 'Grass', 'Electric', 'Ice', 'Fighting',
           'Poison', 'Ground', 'Flying', 'Psychic', 'Bug', 'Rock', 'Ghost',
           'Dark', 'Dragon', 'Steel', 'Fairy']
HEIGHT = 120
WIDTH = 120
CHANNELS = 3
THRESHOLD = 0.5


class PokemonClassificationDataProcessor(InvertibleDataProcessor):
    """Transforms the pokemon dataset at sample_data/pokemon into features and
    labels for classification."""

    def get_raw_features_and_labels(self, dataset_path: str) -> \
            (dict[str, np.ndarray], dict[str, np.ndarray]):
        """Returns the raw feature and label tensors from the dataset path. This
        method is specifically used for the train/val/test sets and not input
        data for prediction, because in some cases the features and labels need
        to be read simultaneously to ensure proper ordering of features and
        labels.

        Raw features are tensors of shape m x h x w x c, where m is the number
        of images, h is the image height, w is the image width, and c is the
        number of channels (3 for RGB), with all values in the interval
        [0, 1]. Raw labels are tensors of shape m x 2, where m is the number of
        examples. All entries are strings from CLASSES indicating 1 or 2 (if
        multi-typed) types belonging to the sample. Types are not ordered.

        :param dataset_path: The path to the file or directory on the local or
            remote filesystem containing the dataset, specifically
            train/val/test and not prediction data.
        :return: A 2-tuple of the features dictionary and labels dictionary,
            with matching keys and ordered tensors.
        """
        # pylint: disable=too-many-locals
        try:
            df = pd.read_csv(os.path.join(dataset_path, LABELS_FILENAME))
        except FileNotFoundError as exc:
            raise LabelsNotFoundError from exc
        X = []
        y = []
        for _, row in df.iterrows():
            name, type1, type2 = row
            full_path = os.path.join(dataset_path,
                                     IMAGES_DIRNAME,
                                     f'{name}.png')
            try:
                # Discard alpha channel.
                tensor = imread(full_path)[:, :, :3]
            except FileNotFoundError:
                # Some pokemon images are jpg--we will not use these.
                tensor = None
            if tensor is not None:
                if not isinstance(type2, str):
                    label = type1, None
                else:
                    label = type1, type2
                X.append(tensor)
                y.append(label)
        X = np.array(X)
        y = np.array(y)
        num_train = int(len(X) * TRAIN_SPLIT)
        num_val = int(len(X) * VAL_SPLIT)
        X_train = X[:num_train]
        y_train = y[:num_train]
        X_val = X[num_train:num_train + num_val]
        y_val = y[num_train:num_train + num_val]
        X_test = X[num_train + num_val:]
        y_test = y[num_train + num_val:]
        return ({'X_train': X_train, 'X_val': X_val, 'X_test': X_test},
                {'y_train': y_train, 'y_val': y_val, 'y_test': y_test})

    def get_raw_features(self, dataset_path: str) -> dict[str, np.ndarray]:
        """Returns the raw feature tensors from the prediction dataset path. Raw
        features are tensors of shape m x h x w x c, where m is the number of
        images, h is the image height, w is the image width, and c is the number
        of channels (3 for RGB), with all values in the interval [0, 1]. The
        features are already scaled because PNG images load into float32 instead
        of uint8.

        :param dataset_path: The path to the file or directory on the local or
            remote filesystem containing the dataset.
        :return: A dictionary whose values are feature tensors and whose
            corresponding keys are the names by which those tensors should be
            referenced. The returned keys will be {'X_train', 'X_val', 'X_test'}
            if the directory indicated by dataset_path ends with 'trainvaltest',
            and {'X_pred'} otherwise.
        """
        X = []
        for filename in os.listdir(os.path.join(dataset_path, IMAGES_DIRNAME)):
            full_path = os.path.join(dataset_path, IMAGES_DIRNAME, filename)
            # Discard alpha channel.
            tensor = imread(full_path)[:, :, :3]
            X.append(tensor)
        X = np.array(X)
        features = {}
        if dataset_path.rstrip('/').endswith('trainvaltest'):
            num_train = int(len(X) * TRAIN_SPLIT)
            num_val = int(len(X) * VAL_SPLIT)
            features['X_train'] = X[:num_train]
            features['X_val'] = X[num_train:num_train + num_val]
            features['X_test'] = X[num_train + num_val:]
        else:
            features['X_pred'] = X
        return features

    def preprocess_features(self, raw_feature_tensor: np.ndarray) -> np.ndarray:
        """Returns the preprocessed feature tensor from the raw tensor. The
        preprocessed features are how training/validation/test as well as
        prediction data are fed into downstream models. The preprocessed tensors
        are of shape m x h x w x c, where m is the number of images, h is the
        image height, w is the image width, and c is the number of channels
        (3 for RGB), with all values in the interval [0, 1].

        :param raw_feature_tensor: The raw features to be preprocessed.
        :return: The preprocessed feature tensor. This tensor is ready for
            downstream model consumption.
        """
        # PNG images, when loaded by imread, are already scaled into [0, 1].
        return raw_feature_tensor.copy()

    def preprocess_labels(self, raw_label_tensor: np.ndarray) -> np.ndarray:
        """Returns the preprocessed label tensor from the raw tensor. The
        preprocessed labels are how training/validation/test as well as
        prediction data are fed into downstream models. Preprocessed labels are
        tensors of shape m x k, where m is the number of examples, and k is the
        number of classes, where each of the k-length vectors are binary,
        multi-label encoded for a minimum of 1 and a maximum of 2 entries per
        vector.

        :param raw_label_tensor: The raw labels to be preprocessed.
        :return: The preprocessed label tensor. This tensor is ready for
            downstream model consumption.
        """
        preprocessed = []
        for label_strs in raw_label_tensor:
            label_arr = PokemonClassificationDataProcessor._get_label_arr(
                *label_strs)
            preprocessed.append(label_arr)
        return np.array(preprocessed)

    def unpreprocess_features(self, feature_tensor: np.ndarray) -> np.ndarray:
        """Returns the raw feature tensor from the preprocessed tensor; inverts
        preprocessing. Improves model interpretability by enabling users to
        transform model inputs into real-world values.

        :param feature_tensor: The preprocessed features to be inverted.
        :return: The raw feature tensor.
        """
        # No preprocessing was necessary.
        return feature_tensor.copy()

    def unpreprocess_labels(self, label_tensor: np.ndarray) -> np.ndarray:
        """Returns the raw label tensor from the preprocessed tensor; inverts
        preprocessing. Improves model interpretability by enabling users to
        transform model outputs into real-world values.

        :param label_tensor: The preprocessed labels to be inverted.
        :return: The raw label tensor.
        """
        unpreprocessed = []
        for label_arr in label_tensor:
            type_strs = PokemonClassificationDataProcessor._get_label_strs(
                label_arr)
            unpreprocessed.append(type_strs)
        return np.array(unpreprocessed)

    @staticmethod
    def _get_label_arr(type1: str, type2: Optional[str]) -> np.ndarray:
        """Returns the label binary array for the given type pair.

        :param type1: The pokemon's first type.
        :param type2: The pokemon's second type, if any (may be None).
        :return: The label for the given type pair; a tensor of shape k where k
            is the number of classes. The tensor will have one 1 if type2 is
            None, otherwise it will have two. The indices of these will be the
            indices of the type names in CLASSES.
        """
        res = [0] * len(CLASSES)
        res[CLASSES.index(type1)] = 1
        if type2:
            res[CLASSES.index(type2)] = 1
        return np.array(res).astype(np.float32)

    @staticmethod
    def _get_label_strs(label_arr: np.ndarray) -> np.ndarray:
        """Returns the label string array for the given type pair.

        :param label_arr: The pokemon's binary label array.
        :return: The string type labels for the given binary label array; a
            tensor of shape 2. The tensor will have one 1 string entry and 1 NA
            if the pokemon is single-typed; otherwise the tensor will have two
            string entries.
        """
        type_strs = []
        for idx in np.where(label_arr == 1)[0]:
            type_strs.append(CLASSES[idx])
        if len(type_strs) == 1:
            type_strs.append(None)
        return np.array(type_strs)

    @staticmethod
    def get_valid_prediction(pred_arr: np.ndarray,
                             threshold: float = THRESHOLD) -> np.ndarray:
        """Returns a valid binary prediction from the raw prediction tensor. A
        valid prediction has one or two 1s, and all other entries are 0. The
        highest value in the prediction array is automatically converted to a 1,
        and the second-highest is converted to a 1 if the value is higher than
        the given decision threshold.

        :param pred_arr: The raw model predictions; a tensor of shape m x k,
            where m is the number of examples and k is the number of classes.
            All entries are in the range [0, 1].
        :param threshold: The decision threshold, in the range [0, 1]. If the
            second-highest value in pred_arr is greater than this threshold, it
            will be converted to a 1. The highest value is automatically
            converted to a 1 (Pokemon have at least 1 type).
        :return: The valid binary predictions; a tensor of shape m x k, where m
            is the number of example and k is the number of classes. All entries
            are in the set {0, 1}, and in each example there are 1 or 2 ones.
        """
        valid_arr = np.zeros_like(pred_arr)
        top_indices = pred_arr.argsort(axis=1)[:, ::-1]
        for idx in range(len(pred_arr)):
            valid_arr[idx, top_indices[idx, 0]] = 1
            if pred_arr[idx, top_indices[idx, 1]] >= threshold:
                valid_arr[idx, top_indices[idx, 1]] = 1
        return valid_arr



"""Contains hashing functions."""

from typing import Collection
import hashlib
from functools import partial
from s3fs import S3FileSystem

CHUNK_SIZE = 2 ** 20


def get_hash_local(files_to_hash: Collection[str]) -> str:
    """Returns the MD5 hex digest string from hashing the content of all the
    given files on the local filesystem. The files are sorted before hashing
    so that the process is reproducible.

    :param files_to_hash: A collection of paths to files whose contents
        should be hashed.
    :return: The MD5 hex digest string from hashing the content of all the
        given files.
    """
    hash_md5 = hashlib.md5()
    for filename in sorted(files_to_hash):
        with open(filename, 'rb') as infile:
            for chunk in iter(partial(infile.read, CHUNK_SIZE), b''):
                hash_md5.update(chunk)
    return hash_md5.hexdigest()


def get_hash_s3(files_to_hash: Collection[str]) -> str:
    """Returns the MD5 hex digest string from hashing the content of all the
    given files in S3. The files are sorted before hashing so that the
    process is reproducible.

    :param files_to_hash: A collection of paths to files whose contents
        should be hashed.
    :return: The MD5 hex digest string from hashing the content of all the
        given files.
    """
    hash_md5 = hashlib.md5()
    fs = S3FileSystem()
    for filename in sorted(files_to_hash):
        with fs.open(filename, 'rb') as infile:
            for chunk in iter(partial(infile.read, CHUNK_SIZE), b''):
                hash_md5.update(chunk)
    return hash_md5.hexdigest()



"""Contains the VersionedModel class."""
# pylint: disable=no-name-in-module

import os
import json
from tempfile import NamedTemporaryFile
from s3fs import S3FileSystem
from tensorflow.keras.models import load_model


class VersionedModel:
    """Represents a versioned model."""

    def __init__(self, path: str) -> None:
        """Instantiates the object.

        :param path: The path, either on the local filesystem or in a cloud
            store such as S3, from which the model should be loaded. An S3 path
            should be a URL of the form "s3://bucket-name/path/to/dir".
        """
        self.path = path
        if path.startswith('s3://'):
            fs = S3FileSystem()
            # Get model.
            with NamedTemporaryFile() as tmp_file:
                with fs.open(os.path.join(path, 'model.h5'), 'rb') as infile:
                    tmp_file.write(infile.read())
                tmp_file.seek(0)
                self.model = load_model(tmp_file.name)
            # Get hash.
            with fs.open(os.path.join(path, 'meta.json'),
                         'r',
                         encoding='utf-8') as infile:
                metadata = json.loads(infile.read())
            self.md5 = metadata['hash']
        else:
            # Get model.
            self.model = load_model(os.path.join(path, 'model.h5'))
            # Get hash.
            with open(os.path.join(path, 'meta.json'),
                      'r',
                      encoding='utf-8') as infile:
                metadata = json.loads(infile.read())
            self.md5 = metadata['hash']

    def __eq__(self, other: 'VersionedModel') -> bool:
        """Returns True if the two objects have the same loaded MD5 hash code,
        False otherwise.

        :param other: The model with which to compare this object.
        :return: True if the object MD5 hashes match.
        """
        return self.md5 == other.md5

    def __hash__(self) -> int:
        """Returns this object's hashcode based on the loaded MD5 hashcode.

        :return: The object's hashcode based on the loaded MD5 hashcode.
        """
        return hash(self.md5)



"""Loads a VersionedModel and uses it to run prediction on unseen data."""

import os
from typing import Collection
import json
import numpy as np
from mlops.dataset.versioned_dataset import VersionedDataset
from mlops.model.versioned_model import VersionedModel
from mlops.examples.image.classification.pokemon_classification_data_processor \
    import PokemonClassificationDataProcessor, DEFAULT_DATASET_PRED_PATH
from mlops.examples.image.classification.publish_dataset import \
    DATASET_PUBLICATION_PATH_LOCAL, DATASET_VERSION
from mlops.examples.image.classification.train_model import \
    MODEL_PUBLICATION_PATH_LOCAL
from mlops.examples.image.classification.errors import NoModelPathsSuppliedError


def model_evaluate(dataset: VersionedDataset,
                   model: VersionedModel) -> float:
    """Returns the model's loss on the test dataset.

    :param dataset: The dataset.
    :param model: The model.
    :return: The model's loss on the test dataset.
    """
    return model.model.evaluate(x=dataset.X_test,
                                y=dataset.y_test)


def model_predict(features: np.ndarray,
                  dataset: VersionedDataset,
                  model: VersionedModel) -> np.ndarray:
    """Returns the model's unpreprocessed predictions on the data located at the
    given path.

    :param features: The preprocessed features on which to run prediction.
    :param dataset: The dataset.
    :param model: The model.
    :return: The model's unpreprocessed predictions on the data located at the
        given path.
    """
    raw_predictions = model.model.predict(features)
    valid_predictions = \
        PokemonClassificationDataProcessor.get_valid_prediction(raw_predictions)
    return dataset.data_processor.unpreprocess_labels(valid_predictions)


def get_best_model(model_paths: Collection[str]) -> VersionedModel:
    """Returns the versioned model with the best performance on the validation
    dataset.

    :param model_paths: The paths to the versioned models to load.
    :return: The versioned model with the best performance on the validation
        dataset.
    """
    if not model_paths:
        raise NoModelPathsSuppliedError
    best_model_path = None
    best_model_val_loss = 0
    for model_path in model_paths:
        meta_path = os.path.join(model_path, 'meta.json')
        with open(meta_path, 'r', encoding='utf-8') as infile:
            contents = json.loads(infile.read())
        model_val_loss = contents['history']['val_loss'][-1]
        if not best_model_path or model_val_loss < best_model_val_loss:
            best_model_path = model_path
            best_model_val_loss = model_val_loss
    return VersionedModel(best_model_path)


def main() -> None:
    """Runs the program."""
    dataset = VersionedDataset(os.path.join(DATASET_PUBLICATION_PATH_LOCAL,
                                            DATASET_VERSION))
    model_filenames = os.listdir(MODEL_PUBLICATION_PATH_LOCAL)
    model_paths = [os.path.join(MODEL_PUBLICATION_PATH_LOCAL, filename)
                   for filename in model_filenames]
    model = get_best_model(model_paths)
    test_err = model_evaluate(dataset, model)
    print(f'Best model\'s test error: {test_err:.3f}')
    features = dataset.data_processor.get_preprocessed_features(
        DEFAULT_DATASET_PRED_PATH)['X_pred']
    predictions = model_predict(features, dataset, model)
    print(f'Predictions:\n{predictions}')


if __name__ == '__main__':
    main()



"""Contains the InvertibleDataProcessor class."""

from abc import abstractmethod
import numpy as np
from mlops.dataset.data_processor import DataProcessor


class InvertibleDataProcessor(DataProcessor):
    """A DataProcessor that can invert any preprocessing to transform
    preprocessed data back into raw, real-world values for analysis and
    interpretability."""

    @abstractmethod
    def unpreprocess_features(self, feature_tensor: np.ndarray) -> np.ndarray:
        """Returns the raw feature tensor from the preprocessed tensor; inverts
        preprocessing. Improves model interpretability by enabling users to
        transform model inputs into real-world values.

        :param feature_tensor: The preprocessed features to be inverted.
        :return: The raw feature tensor.
        """

    @abstractmethod
    def unpreprocess_labels(self, label_tensor: np.ndarray) -> np.ndarray:
        """Returns the raw label tensor from the preprocessed tensor; inverts
        preprocessing. Improves model interpretability by enabling users to
        transform model outputs into real-world values.

        :param label_tensor: The preprocessed labels to be inverted.
        :return: The raw label tensor.
        """



"""Tests training_config.py."""
# pylint: disable=no-name-in-module

from tensorflow.keras.callbacks import History
from mlops.model.training_config import TrainingConfig


def test_training_config_init() -> None:
    """Tests that TrainingConfig objects are initialized correctly."""
    history = History()
    config = TrainingConfig(history, {})
    assert isinstance(config.history, History)
    assert isinstance(config.train_args, dict)



"""Tests versioned_model.py."""
# pylint: disable=redefined-outer-name,no-name-in-module,unused-import

import os
import pytest
from tensorflow.keras.models import Model
from mlops.dataset.versioned_dataset import VersionedDataset
from mlops.model.versioned_model_builder import VersionedModelBuilder
from mlops.model.training_config import TrainingConfig
from mlops.model.versioned_model import VersionedModel
from tests.model.test_versioned_model_builder import \
    _remove_test_directories_local, _remove_test_directories_s3, \
    TEST_MODEL_PUBLICATION_PATH_LOCAL, TEST_MODEL_PUBLICATION_PATH_S3, \
    dataset, model, training_config

EXPECTED_ATTRIBUTES = {'model', 'md5'}


def _publish_test_model_local(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Publishes the versioned model to the local filesystem.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_local()
    builder = VersionedModelBuilder(dataset, model, training_config)
    builder.publish(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1')


def _publish_test_model_s3(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Publishes the versioned model to S3.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _remove_test_directories_s3()
    builder = VersionedModelBuilder(dataset, model, training_config)
    builder.publish(TEST_MODEL_PUBLICATION_PATH_S3, 'v1')


def test_init_loads_model_local(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that the object loads the saved model from the local filesystem.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _publish_test_model_local(dataset, model, training_config)
    model_path = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1')
    versioned_model = VersionedModel(model_path)
    for attribute in EXPECTED_ATTRIBUTES:
        assert hasattr(versioned_model, attribute)


@pytest.mark.awstest
def test_init_loads_model_s3(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that the object loads the saved model from S3.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _publish_test_model_s3(dataset, model, training_config)
    model_path = os.path.join(TEST_MODEL_PUBLICATION_PATH_S3, 'v1')
    versioned_model = VersionedModel(model_path)
    for attribute in EXPECTED_ATTRIBUTES:
        assert hasattr(versioned_model, attribute)


def test_versioned_models_from_same_files_are_equal(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that two versioned models loaded from the same files are considered
    equal in comparisons.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _publish_test_model_local(dataset, model, training_config)
    model_path = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1')
    versioned_model1 = VersionedModel(model_path)
    versioned_model2 = VersionedModel(model_path)
    assert versioned_model1 == versioned_model2


def test_hashcode_is_hash_of_md5_digest(
        dataset: VersionedDataset,
        model: Model,
        training_config: TrainingConfig) -> None:
    """Tests that the hashcode of the model object is the hash of the loaded MD5
    digest.

    :param dataset: The versioned dataset.
    :param model: The model.
    :param training_config: The training configuration.
    """
    _publish_test_model_local(dataset, model, training_config)
    model_path = os.path.join(TEST_MODEL_PUBLICATION_PATH_LOCAL, 'v1')
    versioned_model = VersionedModel(model_path)
    assert hash(versioned_model) == hash(versioned_model.md5)



"""Contains the TrainingConfig class."""
# pylint: disable = no-name-in-module

from typing import Any
from dataclasses import dataclass
from tensorflow.keras.callbacks import History


@dataclass
class TrainingConfig:
    """Contains training configuration and results.

    history: The model's training history.
    train_args: The training arguments.
    """
    history: History
    train_args: dict[str, Any]



"""Contains custom errors."""


class PublicationPathAlreadyExistsError(FileExistsError):
    """Raised when a VersionedDatasetBuilder or VersionedModelBuilder object
    calls its publish method, and the path to which the object is published
    already exists."""


class InvalidDatasetCopyStrategyError(ValueError):
    """Raised when a VersionedDatasetBuilder calls its publish method, and the
    dataset copy strategy is not one of the valid options."""



"""Contains the DataProcessor class."""

from abc import abstractmethod
import numpy as np


class DataProcessor:
    """Transforms a raw dataset into features and labels for downstream model
    training, prediction, etc."""

    def get_preprocessed_features_and_labels(self, dataset_path: str) -> \
            (dict[str, np.ndarray], dict[str, np.ndarray]):
        """Returns the preprocessed feature and label tensors from the dataset
        path. This method is specifically used for the train/val/test sets and
        not input data for prediction, because in some cases the features and
        labels need to be read simultaneously to ensure proper ordering of
        features and labels.

        :param dataset_path: The path to the file or directory on the local or
            remote filesystem containing the dataset, specifically
            train/val/test and not prediction data.
        :return: A 2-tuple of the features dictionary and labels dictionary,
            with matching keys and ordered tensors.
        """
        raw_feature_tensors, raw_label_tensors = \
            self.get_raw_features_and_labels(dataset_path)
        features = {name: self.preprocess_features(raw_feature_tensor)
                    for name, raw_feature_tensor in raw_feature_tensors.items()}
        labels = {name: self.preprocess_labels(raw_label_tensor)
                  for name, raw_label_tensor in raw_label_tensors.items()}
        return features, labels

    def get_preprocessed_features(self, dataset_path: str) -> \
            dict[str, np.ndarray]:
        """Transforms the raw data at the given file or directory into features
        that can be used by downstream models. The data in the directory may be
        the training/validation/test data, or it may be a batch of user data
        that is intended for prediction, or data in some other format.
        Downstream models can expect the features returned by this function to
        be preprocessed in any way required for model consumption.

        :param dataset_path: The path to the file or directory on the local or
            remote filesystem containing the dataset.
        :return: A dictionary whose values are feature tensors and whose
            corresponding keys are the names by which those tensors should be
            referenced. For example, the training features (value) may be called
            'X_train' (key).
        """
        raw_feature_tensors = self.get_raw_features(dataset_path)
        return {name: self.preprocess_features(raw_feature_tensor)
                for name, raw_feature_tensor in raw_feature_tensors.items()}

    @abstractmethod
    def get_raw_features_and_labels(self, dataset_path: str) -> \
            (dict[str, np.ndarray], dict[str, np.ndarray]):
        """Returns the raw feature and label tensors from the dataset path. This
        method is specifically used for the train/val/test sets and not input
        data for prediction, because in some cases the features and labels need
        to be read simultaneously to ensure proper ordering of features and
        labels.

        For example, when handling image data, the raw features would likely be
        tensors of shape m x h x w x c, where m is the number of images, h is
        the image height, w is the image width, and c is the number of channels
        (3 for RGB), with all values in the interval [0, 255]. The raw labels
        may be tensors of shape m, where m is the number of examples, with all
        values in the set {0, ..., k - 1} indicating the class.

        :param dataset_path: The path to the file or directory on the local or
            remote filesystem containing the dataset, specifically
            train/val/test and not prediction data.
        :return: A 2-tuple of the features dictionary and labels dictionary,
            with matching keys and ordered tensors.
        """

    @abstractmethod
    def get_raw_features(self, dataset_path: str) -> dict[str, np.ndarray]:
        """Returns the raw feature tensors from the dataset path. The raw
        features are how training/validation/test as well as prediction data
        enter the data pipeline. For example, when handling image data, the raw
        features would likely be tensors of shape m x h x w x c, where m is the
        number of images, h is the image height, w is the image width, and c is
        the number of channels (3 for RGB), with all values in the interval
        [0, 255].

        :param dataset_path: The path to the file or directory on the local or
            remote filesystem containing the dataset.
        :return: A dictionary whose values are feature tensors and whose
            corresponding keys are the names by which those tensors should be
            referenced. For example, the training features (value) may be called
            'X_train' (key).
        """

    @abstractmethod
    def preprocess_features(self, raw_feature_tensor: np.ndarray) -> np.ndarray:
        """Returns the preprocessed feature tensor from the raw tensor. The
        preprocessed features are how training/validation/test as well as
        prediction data are fed into downstream models. For example, when
        handling image data, the preprocessed features would likely be tensors
        of shape m x h x w x c, where m is the number of images, h is the image
        height, w is the image width, and c is the number of channels (3 for
        RGB), with all values in the interval [0, 1].

        :param raw_feature_tensor: The raw features to be preprocessed.
        :return: The preprocessed feature tensor. This tensor is ready for
            downstream model consumption.
        """

    @abstractmethod
    def preprocess_labels(self, raw_label_tensor: np.ndarray) -> np.ndarray:
        """Returns the preprocessed label tensor from the raw tensor. The
        preprocessed labels are how training/validation/test as well as
        prediction data are fed into downstream models. For example, in a
        classification task, the preprocessed labels may be tensors of shape
        m x k, where m is the number of examples, and k is the number of
        classes, where each of the k-length vectors are one-hot encoded.

        :param raw_label_tensor: The raw labels to be preprocessed.
        :return: The preprocessed label tensor. This tensor is ready for
            downstream model consumption.
        """
